class Error(Exception):
    def __init__(self, code, message):
        super(Error, self).__init__(message)
        self.code = code
        self.message = message


class InvalidEntityIngestLink(Error):
    def __init__(self, from_entity, to_entity):
        message = f'It is not possible to link a {from_entity.type} to {to_entity.type} in ingest database.'
        super(InvalidEntityIngestLink, self).__init__('InvalidEntityIngestLink', message)
        self.from_entity = from_entity
        self.to_entity = to_entity


class InvalidLinkInSpreadsheet(Error):
    def __init__(self, from_entity, link_entity_type, link_entity_id):
        message = f'It is not possible to link a {from_entity.type} to {link_entity_type} in the spreadsheet.'
        super(InvalidLinkInSpreadsheet, self).__init__('InvalidLinkInSpreadsheet', message)
        self.from_entity = from_entity
        self.link_entity_type = link_entity_type
        self.link_entity_id = link_entity_id


class LinkedEntityNotFound(Error):
    def __init__(self, from_entity, entity_type, id):
        message = f'A link from a {from_entity.type} with id {from_entity.id} to a {entity_type} with id, ' \
                  f'"{id}", is not found in the spreadsheet.'
        super(LinkedEntityNotFound, self).__init__('LinkedEntityNotFound', message)
        self.entity = entity_type
        self.id = id


class MultipleProcessesFound(Error):
    def __init__(self, from_entity, process_ids):
        message = f'Multiple processes are linked {from_entity.type} in the spreadsheet: {process_ids}.'
        super(MultipleProcessesFound, self).__init__('MultipleProcessesFound', message)

        self.process_ids = process_ids
        self.from_entity = from_entity



import copy
import json
import logging

from hca_ingest.api.ingestapi import IngestApi
from hca_ingest.importer.submission.entity import Entity
from hca_ingest.importer.submission.entity_map import EntityMap
from hca_ingest.importer.submission.submission import Submission

ENTITY_LINK = {
    'biomaterial': 'biomaterials',
    'process': 'processes',
    'file': 'files',
    'protocol': 'protocols',
    'project': 'projects',
    'submission_envelope': 'submissionEnvelopes'
}


def json_equals(json1: dict, json2: dict):
    return json.dumps(json1, sort_keys=True) == json.dumps(json2, sort_keys=True)


class IngestSubmitter(object):
    def __init__(self, ingest_api: IngestApi):
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)
        self.PROGRESS_CTR = 50
        self.logger = logging.getLogger(__name__)

    def add_entity(self, entity: Entity, submission_url: str):
        link_name = ENTITY_LINK[entity.type]

        if entity.type == 'file':
            file_name = entity.content['file_core']['file_name']
            response = self.ingest_api.create_file(submission_url,
                                                   file_name,
                                                   entity.content)
        elif entity.type == 'project':
            response = self.ingest_api.create_project(submission_url,
                                                      entity.content)
        else:
            response = self.ingest_api.create_entity(submission_url,
                                                     {"content": entity.content},
                                                     link_name)
        entity.ingest_json = response

        return entity

    def add_entities(self, entity_map: EntityMap, submission_url: str) -> Submission:
        submission = Submission(self.ingest_api, submission_url)
        submission.define_manifest(entity_map)
        for e in entity_map.get_new_entities():
            self.add_entity(e, submission_url)
            submission.add_entity(e)

        return submission

    def update_entities(self, entity_map: EntityMap):
        updated_entities = [self.update_entity(e) for e in entity_map.get_entities() if e.is_reference]
        return updated_entities

    def update_entity(self, entity: Entity):
        if not entity.ingest_json:
            entity.ingest_json = self.ingest_api.get_entity_by_uuid(ENTITY_LINK[entity.type], entity.id)

        patch = copy.deepcopy(entity.ingest_json.get('content'))
        patch.update(entity.content)
        if not json_equals(entity.ingest_json.get('content'), patch):
            self.ingest_api.patch(entity.url, json={'content': patch})

        return entity

    def link_submission_to_project(self, project_uuid: str, submission_url: str):
        project_entity = self.ingest_api.get_entity_by_uuid('projects', project_uuid)
        submission_envelope = self.ingest_api.get_submission(submission_url)
        self.ingest_api.link_entity(project_entity, submission_envelope, 'submissionEnvelopes')

    def link_entity(self, from_entity: Entity, to_entity: Entity, relationship: str, is_collection=True):
        if from_entity.is_linking_reference and not from_entity.ingest_json:
            from_entity.ingest_json = self.ingest_api.get_entity_by_uuid(ENTITY_LINK[from_entity.type],
                                                                         from_entity.id)

        if to_entity.is_linking_reference and not to_entity.ingest_json:
            to_entity.ingest_json = self.ingest_api.get_entity_by_uuid(ENTITY_LINK[to_entity.type], to_entity.id)

        from_entity_ingest = from_entity.ingest_json
        to_entity_ingest = to_entity.ingest_json

        self.ingest_api.link_entity(from_entity_ingest, to_entity_ingest, relationship, is_collection)

    def link_entities(self, entity_map: EntityMap, submission: Submission):
        progress = 0
        for entity in entity_map.get_entities():
            for link in entity.direct_links:
                to_entity = entity_map.get_entity(link['entity'], link['id'])
                try:
                    self.link_entity(entity, to_entity, relationship=link['relationship'],
                                     is_collection=link.get('is_collection', True))
                    progress = progress + 1
                    expected_links = int(submission.manifest.get('expectedLinks', 0))
                    if progress % self.PROGRESS_CTR == 0 or (progress == expected_links):
                        manifest_url = self.ingest_api.get_link_from_resource(submission.manifest, 'self')
                        self.ingest_api.patch(manifest_url, json={'actualLinks': progress})
                        self.logger.info(f"links progress: {progress}/ {submission.manifest.get('expectedLinks')}")

                except Exception as link_error:
                    error_message = f'''The {entity.type} with id {entity.id} could not be linked to {to_entity.type} \
                    with id {to_entity.id}.'''
                    self.logger.error(error_message)
                    self.logger.error(f'{str(link_error)}')
                    raise



import copy
import logging

from openpyxl.worksheet.worksheet import Worksheet

from hca_ingest.api.ingestapi import IngestApi
from hca_ingest.importer.conversion import conversion_strategy
from hca_ingest.importer.conversion.column_specification import ColumnSpecification
from hca_ingest.importer.conversion.conversion_strategy import CellConversion
from hca_ingest.importer.conversion.metadata_entity import MetadataEntity
from hca_ingest.importer.data_node import DataNode
from hca_ingest.importer.spreadsheet.ingest_worksheet import IngestWorksheet, \
    MODULE_TITLE_PATTERN, IngestRow
from hca_ingest.template.exceptions import UnknownKeySchemaException
from hca_ingest.template.schema_template import SchemaTemplate


class TemplateManager:
    default_keys = ['describedBy', 'schema_type']

    def __init__(self, template: SchemaTemplate, ingest_api: IngestApi):
        self.template = template
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)

    def create_template_node(self, worksheet: Worksheet):
        concrete_entity = self.get_concrete_type(worksheet.title)
        schema = self._get_schema(concrete_entity)
        data_node = DataNode()
        data_node[self.default_keys[0]] = schema['url']
        data_node[self.default_keys[1]] = schema['domain_entity']
        return data_node

    def create_row_template(self, ingest_worksheet: IngestWorksheet):
        concrete_type = self.get_concrete_type(ingest_worksheet.title)
        domain_type = self.get_domain_type(concrete_type)
        column_headers = ingest_worksheet.get_column_headers()
        cell_conversions = []

        context = self._determine_context(concrete_type, ingest_worksheet)
        header_counter = {}
        for header in column_headers:
            if not header_counter.get(header):
                header_counter[header] = 0
            header_counter[header] = header_counter[header] + 1

            column_spec = ColumnSpecification(self.template, header, concrete_type, context=context,
                                              order_of_occurrence=header_counter[header])
            strategy = conversion_strategy.determine_strategy(column_spec)
            cell_conversions.append(strategy)

        default_values = None
        if not ingest_worksheet.is_module_tab():
            default_values = self._define_default_values(concrete_type)

        return RowTemplate(domain_type, concrete_type, cell_conversions,
                           default_values=default_values)

    @staticmethod
    def _determine_context(concrete_type, ingest_worksheet):
        context_components = [concrete_type]
        module_field_name = ingest_worksheet.get_module_field_name()
        if module_field_name:
            context_components.append(module_field_name)
        context = '.'.join(context_components)
        return context

    def _define_default_values(self, object_type):
        default_values = {
            self.default_keys[0]: self.get_schema_url(object_type),
            self.default_keys[1]: self.get_domain_type(object_type)
        }
        return default_values

    def get_latest_schema_url(self, high_level_entity, domain_entity, concrete_entity):
        latest_schema = self.ingest_api.get_schemas(
            latest_only=True,
            high_level_entity=high_level_entity,
            domain_entity=domain_entity.split('/')[0],
            concrete_entity=concrete_entity
        )

        return latest_schema[0]['_links']['json-schema']['href'] if latest_schema else None

    def get_schema_url(self, concrete_entity):
        schema = self._get_schema(concrete_entity)
        # TODO use schema version that is specified in spreadsheet for now
        return schema.get('url') if schema else None

    # TODO this just 2 lines. Perhaps we can just inline this to client code?
    def _get_schema(self, concrete_entity):
        spec = self.lookup(concrete_entity)
        return spec.get('schema') if spec else None

    def get_concrete_type(self, title):
        """
        Concrete Type refers to the specific type of an object based on a given schema.
        This method determines the concrete type given the worksheet title.

        :param title: the title of the worksheet.
        :return: the Concrete Type of a given worksheet title
        """
        result = MODULE_TITLE_PATTERN.search(title)
        if not result:
            raise InvalidTabName(title)
        main_label = result.group('main_label')
        return self.template.lookup_metadata_schema_name_given_title(main_label)

    def get_domain_type(self, concrete_type):
        """
        Domain Entity Type is the high level classification of Concrete Entities. For example,
        Donor Organism belongs to the Biomaterial domain; all Donor Organisms are considered
        Biomaterials.

        :param concrete_type: the actual metadata entity type
        :return: the Domain Entity Type for the given concrete_entity
        """
        domain_type = None
        spec = self.lookup(concrete_type)
        schema = spec.get('schema') if spec else None
        if schema:
            domain_type = schema.get('domain_entity', '')
            type_components = domain_type.split('/')
            if type_components:
                domain_type = type_components[0]
        return domain_type

    # Created this convenience method so that clients don't have to call template manager twice.
    # Unlike _get_schema, this will be used outside the context of this class.
    def get_worksheet_domain_type(self, title):
        concrete_type = self.get_concrete_type(title)
        return self.get_domain_type(concrete_type)

    def lookup(self, header_name):
        try:
            spec = self.template.lookup_property_from_template(header_name)
        except UnknownKeySchemaException:
            self.logger.warning(f'UnknownKeySchemaException: Could not lookup {header_name} in template.')
            return {}

        return spec

    def get_schemas(self):
        return self.template.metadata_schema_urls


def build(schemas, ingest_api) -> TemplateManager:
    if not schemas:
        template = SchemaTemplate(ingest_api_url=ingest_api.url)
    else:
        template = SchemaTemplate(ingest_api_url=ingest_api.url, metadata_schema_urls=schemas)

    template_mgr = TemplateManager(template, ingest_api)
    return template_mgr


class RowTemplate:
    def __init__(self, domain_type, object_type, cell_conversions, default_values=None):
        self.domain_type = domain_type
        self.concrete_type = object_type
        self.cell_conversions = cell_conversions
        self.default_values = copy.deepcopy(default_values) if default_values else {}

    def do_import(self, row: IngestRow, is_module=False):
        row_errors = []
        metadata = MetadataEntity(domain_type=self.domain_type, concrete_type=self.concrete_type,
                                  content=self.default_values, row=row, is_module=is_module)
        for index, cell in enumerate(row.values):
            if cell.value is None:
                continue
            conversion: CellConversion = self.cell_conversions[index]
            try:
                conversion.apply(metadata, cell.value)
            except ValueError as e:
                error_detail = f'Could not convert to {conversion.converter.type().value} ' \
                               f'from: {type(cell.value).__name__} {cell.value}'
                if type(cell.value) == float:
                    error_detail += " This may be because the default excel number format is float."

                row_errors.append({
                    "location": f'column={index}, value={cell.value}',
                    "type": "Conversion Error",
                    "detail": error_detail
                })
            except Exception as e:
                row_errors.append({
                    "location": f'column={index}, value={cell.value}',
                    "type": e.__class__.__name__,
                    "detail": str(e)
                })
        return metadata, row_errors


class ParentFieldNotFound(Exception):
    def __init__(self, header_name):
        message = f'{header_name} has no parent field'
        super(ParentFieldNotFound, self).__init__(message)

        self.header_name = header_name


class InvalidTabName(Exception):

    def __init__(self, tab_name):
        super(InvalidTabName, self).__init__(f'Invalid tab name [{tab_name}]')
        self.tab_name = tab_name



from datetime import datetime

import requests
import yaml

from hca_ingest.api.ingestapi import IngestApi
from hca_ingest.template.descriptor import SimplePropertyDescriptor
from .exceptions import RootSchemaException, UnknownKeySchemaException
from .migration_parser import MigrationParser
from .schema_parser import SchemaParser
from .tab_config import TabConfig

EXTERNAL_REFERENCE_FIELD = 'uuid'


class SchemaTemplate():
    """ A SchemaTemplate is used to encapsulate information about all the metadata schema files and
    property migration files that are directly passed in in order to generate a spreadsheet. """

    # TODO: simplify construction, move existing opinionated constructor to a static method or builder
    def __init__(self, ingest_api_url="http://api.ingest.dev.archive.data.humancellatlas.org",
                 migrations_url="https://schema.humancellatlas.org/property_migrations",
                 metadata_schema_urls=None, json_schema_docs=None, tab_config=None, property_migrations=None,
                 custom_properties=None):
        """ Creates and empty/default dictionary containing the following information:
        1) template_version:  A string keeping track of the version of the TopLevelSchemaDescriptor that is being used
        in case the components of this dictionary changes over time.

        2) created_date: A string representing the date and time at which point this internal representation of the
        metadata schemas and property migration file was created.

        3) meta_data_properties: A dictionary structure where each key represents a metadata schema that was directly
        passed in to parse and where each value is a dictionary representing any properties that are associated with
        the given schema. The dictionary itself may be recursive and contain embedded schemas that describe entire
        properties.

        4) labels: A dictionary structure mapping the "path" to every property that is accessible by traversing the
        metadata schema graph structure, initiating from this top-level schema. For example, the "donor_organism" Type
        metadata schema contains the property "biomaterial_core" which is described by its own Core schema and contains
        a property "biomaterial_name". The key that would be added here would be "biomaterial_name" and its value would
        be "donor_organism.biomaterial_core.biomaterial_name".

        5) tabs: A list of dictionaries where each dictionary represents a tab that would be generated in the overall
        spreadsheet if this entire schema was turned into a spreadsheet.

        6) migrations: A nested dictionary describing the migrations that have occurred for each property in any
        metadata schema.

        :param ingest_api_url: A string representing the root Ingest API URL from which the latest metadata schema
                               URLs will be queried.
        :param migrations_url: A string representing the URL containing the property migration file that enumerates
                               changes in metadata schemas between versions.
        :param metadata_schema_urls: A list of strings representing URLs, each containing a JSON-formatted metadata
                                     schema. If this parameter is provided, the parameter json_schema_docs should not be
                                     populated.
        :param json_schema_docs: A list of objects where each object represents a deserialized JSON-formatted
                                 metadata schema. If this parameter is provided, the parameter metadata_schema_urls
                                 should not be populated.
        :param spreadsheet_configuration: A TabConfig object that represents the expected tabs that will be created
                                          in a spreadsheet given the metadata schemas provided.
        :param property_migrations: An object representing a deserialized JSON-string which is a property migrations
                                    file. A property migrations files dictates how the version of the schema has changed
                                    from an older version.
        :param custom_properties: An object representing a deserialized JSON-string which contains custom fields added
                                  on top of metadata schema
        """

        # Function validation: only one of json_schema_docs or metadata_schema_urls may be populated or neither (NAND
        # boolean).
        if bool(metadata_schema_urls) and bool(json_schema_docs):
            raise Exception(
                'Only one of function arguments metadata_schema_urls or json_schema_docs (or neither) may be '
                'populated when initializing SchemaTemplate.')

        self.metadata_schema_urls = metadata_schema_urls
        # If neither the metadata_schema_urls are given nor the json_schema_docs, fetch the metadata schema URLs via
        # querying the Ingest API.
        if not metadata_schema_urls and not json_schema_docs:
            self.metadata_schema_urls = self._get_latest_submittable_schema_urls(ingest_api_url)

        self.property_migrations = property_migrations
        # If the property migrations were not given as input, read the migrations from the migrations URL and store
        # into the deserialized JSON into a Python object.
        if not self.property_migrations:
            self.property_migrations = self._get_property_migrations_json_obj(migrations_url)

        self.json_schemas = json_schema_docs
        if not self.json_schemas:
            self.json_schemas = self._get_json_objs_from_metadata_schema_urls()
        else:
            self.metadata_schema_urls = [schema.get("$id") or schema.get("id") for schema in self.json_schemas]

        self.template_version = "1.0.0"
        self.created_date = str(datetime.now())
        self.meta_data_properties = {}
        self.custom_properties = {} if not custom_properties else custom_properties
        self.labels = {}
        self.tabs = []
        self.migrations = {}

        for json_schema in self.json_schemas:
            schema_parser = SchemaParser(json_schema)
            schema_descriptor = schema_parser.schema_descriptor
            fully_parsed_dictionary = schema_parser.schema_dictionary
            self.meta_data_properties[schema_descriptor.get_schema_module_name()] = fully_parsed_dictionary
            self.tabs.append(schema_parser.get_tab_representation_of_schema())
            self.labels.update(schema_parser.get_map_of_paths_by_property_label(
                {schema_descriptor.get_schema_module_name(): fully_parsed_dictionary}))
            self.init_custom_properties(schema_descriptor)

        self.migrations = MigrationParser(self.property_migrations).migrations

        self.spreadsheet_configuration = tab_config if tab_config else TabConfig(self.get_dictionary_representation())

    def init_custom_properties(self, schema_descriptor):
        external_field_descriptor = SimplePropertyDescriptor({})
        external_field_descriptor.identifiable = True
        external_field_descriptor.external_reference = True
        self.custom_properties[schema_descriptor.get_schema_module_name()] = {}
        self.custom_properties[schema_descriptor.get_schema_module_name()][
            EXTERNAL_REFERENCE_FIELD] = external_field_descriptor.get_dictionary_representation_of_descriptor()

    def get_dictionary_representation(self):
        return {
            "template_versions": self.template_version,
            "created_date": self.created_date,
            "meta_data_properties": self.meta_data_properties,
            "labels": self.labels,
            "tabs": self.tabs,
            "migrations": MigrationParser.get_dictionary_representation(self.migrations)
        }

    def get_list_of_schema_spreadsheet_representations(self):
        return self.tabs

    def lookup_property_attributes_in_metadata(self, property_key):
        """
        Given a property key which details the full path to the property from the top level schema, returns a dictionary
        representing the attributes of the requested property.

        :param property_key: A string representing the fully qualified path to the desired metadata property
        :return: A dictionary representing the attributes of the property. If none is found, an exception will be thrown
        """

        result = self._lookup_fully_qualified_key_path_in_dictionary(property_key, self.meta_data_properties)

        if not result:
            raise UnknownKeySchemaException(f"ERROR: Cannot find key {property_key} in any schema!")

        return result

    def lookup_property_from_template(self, property_key):
        """
        Given a property key which details the full path to the property from the top level schema, returns a dictionary
        representing the attributes of the requested property.

        :param property_key: A string representing the fully qualified path to the desired metadata or custom property
        :return: A dictionary representing the attributes of the property. If none is found, an exception will be thrown
        """
        result = self._lookup_fully_qualified_key_path_in_dictionary(property_key, self.meta_data_properties)

        if not result:
            result = self._lookup_fully_qualified_key_path_in_dictionary(property_key, self.custom_properties)

        if not result:
            raise UnknownKeySchemaException(f"ERROR: Cannot find key {property_key} in any schema!")

        return result

    def lookup_metadata_schema_name_given_title(self, tab_display_name):
        """
        Given a tab's display name which is often used to label a column in a spreadsheet that represents a property of
        a schema, return the fully qualified path that is the equivalent key (the key can then be used to fetch
        additional attributes about the property).

        :param tab_display_name: A string representing a display name of a schema property.
        :return: A string representing the display name's fully qualified path from a schema.
        """

        try:
            return self.spreadsheet_configuration.get_key_for_label(tab_display_name)
        except KeyError:
            raise UnknownKeySchemaException(
                f"ERROR: Was unable to find a fully qualified path (key) with the display name {tab_display_name}.")

    def lookup_next_latest_key_migration(self, fully_qualified_key):
        """
        Given a key, returns the fully qualified key in the next latest schema (potentially not the absolute latest
        schema though) if it has been replaced.

        :param key: A string representing a fully qualified field name in an older metadata schema version.
        :return: A string representing a fully qualified field name in the next latest metadata schema version if a
                 migration to that schema version exists. If a migration does not exist, returns the original fully
                 qualified key.
        """

        if not self._validate_fully_qualified_key_exists(fully_qualified_key):
            raise Exception(f"ERROR: Fully qualified key {fully_qualified_key} does not exist in any of the schemas.")

        if fully_qualified_key in self.migrations:
            migration_info = self.migrations[fully_qualified_key]
            return migration_info.replaced_by if migration_info.replaced_by in migration_info else fully_qualified_key

        return fully_qualified_key

    def lookup_absolute_latest_key_migration(self, fully_qualified_key):
        """
        Given a key, returns the fully qualified key in the maximal latest schema if it has been replaced.

        :param key: A string representing a fully qualified field name in an older metadata schema version.
        :return: A string representing a fully qualified field name in the latest metadata schema
                 version if a migration chain to that schema version exists. If a migration does not exist, returns the
                 original fully qualified key.
        """

        if not self._validate_fully_qualified_key_exists(fully_qualified_key):
            raise Exception(f"ERROR: Fully qualified key {fully_qualified_key} does not exist in any of the schemas.")

        original_key = fully_qualified_key
        next_latest_migrated_key = self.lookup_next_latest_key_migration(fully_qualified_key)

        while original_key != next_latest_migrated_key:
            original_key = next_latest_migrated_key
            next_latest_migrated_key = self.lookup_next_latest_key_migration(original_key)

        return next_latest_migrated_key

    def generate_yaml_representation_of_spreadsheets(self, tabs_only):
        """
        Generate a YAML object representing the schema_template object and if `tabs_only` is true, then only generate
        the YAML based on self.tabs. This YAML object can be used to generate arbitrary spreadsheets based on the
        metadata schemas.

        :param tabs_only: When true, only uses the tabs to generate the YAML instead of including all the "metadata"
                          about the metadata schemas encapsulated in this SchemaTemplate object.
        :return: A YAML object representing the metadata schemas encapsulated in this SchemaTemplate object.
        """

        yaml.default_flow_style = False
        return yaml.dump({"tabs": self.tabs} if tabs_only else self.get_dictionary_representation(), indent=4)

    def _validate_fully_qualified_key_exists(self, fully_qualified_key):
        return fully_qualified_key in self.labels

    def _lookup_fully_qualified_key_path_in_dictionary(self, fully_defined_metadata_property, dictionary):
        if fully_defined_metadata_property in dictionary:
            return dictionary[fully_defined_metadata_property]

        if '.' in fully_defined_metadata_property:
            expected_schema = fully_defined_metadata_property.split('.')[0]
            if expected_schema in dictionary:
                return self._lookup_fully_qualified_key_path_in_dictionary(
                    fully_defined_metadata_property.split('.', 1)[1], dictionary[expected_schema])

    def _get_latest_submittable_schema_urls(self, ingest_api_url):
        """
        Given the Ingest API URL, send a request to get the latest metadata schemas and return the URLs.

        :param ingest_api_url: A string representing the API URL from which to query for schemas.
        :return: A list of strings where each string represents a URL representing the location of a JSON-formatted
                 metadata schema.
        """

        ingest_api = IngestApi(url=ingest_api_url)
        raw_schemas_from_ingest_api = ingest_api.get_schemas(high_level_entity="type", latest_only=True)
        return [schema["_links"]["json-schema"]["href"] for schema in raw_schemas_from_ingest_api]

    def _get_property_migrations_json_obj(self, migrations_url):
        """
        Returns a a JSON-formatted property migrations file read from the given URL.

        :param migrations_url: A string representing the URL from which to read the JSON-formatted property
                               migrations file.
        :return: An object representing the deserialized JSON-formatted property migrations file.
        """

        try:
            return requests.get(migrations_url).json()["migrations"]
        except Exception:
            raise RootSchemaException(f"Was unable to read the property migrations file from URL {migrations_url}")

    def _get_json_objs_from_metadata_schema_urls(self):
        """
        Return a list of objects that represent deserialized JSON-formatted metadata schemas from the URLs stored in
        self.metadata_schema_urls.

        :return: A list of objects representing deserialized JSON-formatted metadata schemas.
        """
        metadata_schema_objs = []
        for uri in self.metadata_schema_urls:
            try:
                metadata_schema_objs.append(requests.get(url=uri).json())
            except Exception:
                raise RootSchemaException(f"Was unable to read metadata schema JSON at {uri}")
        return metadata_schema_objs



from itertools import groupby
from typing import List

from hca_ingest.downloader.entity import Entity
from hca_ingest.importer.spreadsheet.ingest_workbook import SCHEMAS_WORKSHEET

MODULE_WORKSHEET_NAME_CONNECTOR = ' - '
SCALAR_LIST_DELIMITER = '||'

ONTOLOGY_PROPS = ['ontology', 'ontology_label', 'text']
EXCLUDE_KEYS = ['describedBy', 'schema_type']


class Flattener:
    def __init__(self):
        self.workbook = {}
        self.schemas = {}

    def flatten(self, entity_list: List[Entity]):
        for entity in entity_list:
            if entity.concrete_type != 'process':
                self._flatten_entity(entity)
            self._extract_schema_url(entity.content, entity.concrete_type)

        self.workbook[SCHEMAS_WORKSHEET] = list(self.schemas.values())
        return self.workbook

    def _flatten_entity(self, entity: Entity):
        worksheet_name = entity.concrete_type
        row = {f'{worksheet_name}.uuid': entity.uuid}

        if not worksheet_name:
            raise ValueError('There should be a worksheet name')

        self._flatten_object(entity.content, row, parent_key=worksheet_name)

        if entity.input_biomaterials or entity.input_files:
            embedded_content = self.embed_link_columns(entity)
            self._flatten_object(embedded_content, row)

        self._add_row_to_worksheet(row, worksheet_name)

    def _flatten_module_list(self, module_list: dict, object_key: str):
        for module in module_list:
            self._flatten_module(module, object_key)

    def _flatten_module(self, obj: dict, object_key: str):
        worksheet_name = object_key
        module_row = {}

        if not worksheet_name:
            raise ValueError('There should be a worksheet name')

        self._flatten_object(obj, module_row, parent_key=worksheet_name)

        self._add_row_to_worksheet(module_row, worksheet_name)

    def _add_row_to_worksheet(self, row, worksheet_name):
        user_friendly_worksheet_name = self._format_worksheet_name(worksheet_name)
        worksheet = self.workbook.get(user_friendly_worksheet_name, {'headers': [], 'values': []})
        rows = worksheet.get('values')
        rows.append(row)
        headers = self._update_headers(row, worksheet)
        self.workbook[user_friendly_worksheet_name] = {
            'headers': headers,
            'values': rows
        }

    def embed_link_columns(self, entity: Entity):
        embedded_content = {}
        self._embed_process(entity, embedded_content)
        self._embed_protocol_ids(entity, embedded_content)
        self._embed_input_ids(entity, embedded_content)
        return embedded_content

    def _embed_process(self, entity: Entity, embedded_content):
        embed_process = {
            'process': {
                'uuid': entity.process.uuid
            }
        }
        embed_process['process'].update(entity.process.content)
        embedded_content.update(embed_process)

    def _embed_protocol_ids(self, entity: Entity, embedded_content):
        protocols_by_type = {}
        for p in entity.protocols:
            p: Entity
            protocols = protocols_by_type.get(p.concrete_type, [])
            protocols.append(p)
            protocols_by_type[p.concrete_type] = protocols

        for concrete_type, protocols in protocols_by_type.items():
            protocol_ids = [p.content['protocol_core']['protocol_id'] for p in protocols]
            protocol_uuids = [p.uuid for p in protocols]
            embedded_protocol_ids = {
                concrete_type: {
                    'protocol_core': {
                        'protocol_id': protocol_ids
                    },
                    'uuid': protocol_uuids
                }
            }
            embedded_content.update(embedded_protocol_ids)

    def _embed_input_ids(self, entity: Entity, embedded_content):
        self._embed_input_biomaterial_ids(embedded_content, entity)
        self._embed_input_file_ids(embedded_content, entity)

    def _embed_input_biomaterial_ids(self, embedded_content, entity):
        for concrete_type, inputs_iter in groupby(entity.input_biomaterials, lambda e: e.concrete_type):
            inputs = list(inputs_iter)
            input_ids_ids = [i.content['biomaterial_core']['biomaterial_id'] for i in inputs]
            input_ids_uuids = [i.uuid for i in inputs]
            embedded_input_ids = {
                concrete_type: {
                    'biomaterial_core': {
                        'biomaterial_id': input_ids_ids
                    },
                    'uuid': input_ids_uuids
                }
            }
            embedded_content.update(embedded_input_ids)

    def _embed_input_file_ids(self, embedded_content, entity):
        for concrete_type, inputs_iter in groupby(entity.input_files, lambda e: e.concrete_type):
            inputs = list(inputs_iter)
            input_ids_ids = [i.content['file_core']['file_name'] for i in inputs]
            input_ids_uuids = [i.uuid for i in inputs]
            embedded_input_ids = {
                concrete_type: {
                    'file_core': {
                        'file_name': input_ids_ids
                    },
                    'uuid': input_ids_uuids
                }
            }
            embedded_content.update(embedded_input_ids)

    def _extract_schema_url(self, content: dict, concrete_entity: str):
        schema_url = content.get('describedBy')
        existing_schema_url = self.schemas.get(concrete_entity)
        self._validate_no_schema_version_conflicts(existing_schema_url, schema_url)

        if not existing_schema_url:
            self.schemas[concrete_entity] = schema_url

    def _validate_no_schema_version_conflicts(self, existing_schema_url, schema_url):
        if existing_schema_url and existing_schema_url != schema_url:
            raise ValueError(f'The concrete entity schema version should be consistent across entities.\
                    Multiple versions of same concrete entity schema is found:\
                     {schema_url} and {existing_schema_url}')

    def _update_headers(self, row, worksheet):
        headers = worksheet.get('headers')
        for key in row.keys():
            if key not in headers:
                headers.append(key)
        return headers

    def _flatten_object(self, object: dict, flattened_object: dict, parent_key: str = ''):
        if isinstance(object, dict):
            for key in object:
                if key in EXCLUDE_KEYS:
                    continue

                value = object[key]
                full_key = f'{parent_key}.{key}' if parent_key else key
                if isinstance(value, dict) or isinstance(value, list):
                    self._flatten_object(value, flattened_object, parent_key=full_key)
                else:
                    flattened_object[full_key] = str(value)
        elif isinstance(object, list):
            self._flatten_list(flattened_object, object, parent_key)

    def _flatten_list(self, flattened_object, object, parent_key):
        if self._is_list_of_objects(object):
            self._flatten_object_list(flattened_object, object, parent_key)
        else:
            self._flatten_scalar_list(flattened_object, object, parent_key)

    def _flatten_scalar_list(self, flattened_object, object, parent_key):
        stringified = [str(e) for e in object]
        flattened_object[parent_key] = SCALAR_LIST_DELIMITER.join(stringified)

    def _flatten_object_list(self, flattened_object: dict, object: dict, parent_key: str):
        if self._is_list_of_ontology_objects(object):
            self._flatten_to_include_object_list_to_main_entity_worksheet(object, flattened_object, parent_key)
        elif self._is_project(parent_key):
            self._flatten_module_list(object, parent_key)
        else:
            self._flatten_to_include_object_list_to_main_entity_worksheet(object, flattened_object, parent_key)

    def _flatten_to_include_object_list_to_main_entity_worksheet(self, object: dict, flattened_object: dict,
                                                                 parent_key: str):
        keys = self._get_keys_of_a_list_of_object(object)

        for key in keys:
            flattened_object[f'{parent_key}.{key}'] = SCALAR_LIST_DELIMITER.join(
                [elem.get(key) for elem in object
                 if elem.get(key) is not None and elem.get(key) != ''])

    def _format_worksheet_name(self, worksheet_name):
        names = worksheet_name.split('.')
        names = [n.replace('_', ' ') for n in names]
        new_worksheet_name = MODULE_WORKSHEET_NAME_CONNECTOR.join([n.capitalize() for n in names])
        return new_worksheet_name

    def _is_list_of_objects(self, content):
        return content and isinstance(content[0], dict)

    def _is_list_of_ontology_objects(self, object: dict):
        first_elem = object[0] if object else {}
        result = [prop in first_elem for prop in ONTOLOGY_PROPS]
        # TODO better check the schema if field is ontology
        return any(result)

    def _get_keys_of_a_list_of_object(self, object: dict):
        first_elem = object[0] if object else {}
        return list(first_elem.keys())

    def _is_project(self, parent_key: str):
        entity_type = parent_key.split('.')[0]
        return entity_type == 'project'



#!/usr/bin/env python
"""
Class encapsulating implementation details on the Descriptor classes. Descriptors represent a portion of a metadata
schema.
"""

import re

IDENTIFIABLE_PROPERTIES = ["biomaterial_id", "process_id", "protocol_id", "file_name"]


class Descriptor():
    """ Parent class type. A Descriptor type encapsulate a small isolated amount of information about a portion of a
    metadata schema.
    """

    def get_dictionary_representation_of_descriptor(self):
        """ Returns a dict representing the Descriptor object. """
        raise NotImplementedError("Subclasses of Descriptor are required to override this method.")


class SchemaTypeDescriptor(Descriptor):
    """ Descriptor encapsulating "metadata" information about a single metadata schema file. """

    def __init__(self, metadata_schema_url):
        url_validation_regex = re.compile(
            r'^http[s]?://(?P<location>([^/]+/)*[^/]+)/' +
            r'(?P<high_level_entity>(type)|(module)|(core)|(system))/' +
            r'((?P<domain_entity>([^/]+/)*[^/]+)/)?' +
            r'(?P<version>(?P<version_number>(?P<major>\d+)(\.(?P<minor>\d+))?(\.(?P<rev>\d+))?)|(?P<latest>latest))/' +
            r'(?P<module>.*)$'
        )
        if not url_validation_regex.match(metadata_schema_url):
            raise Exception(
                f"ERROR: The metadata schema URL passed in for parsing {metadata_schema_url} does not conform to "
                f"expected format.")

        self.high_level_entity = url_validation_regex.match(metadata_schema_url).group("high_level_entity")
        self.domain_entity = url_validation_regex.match(metadata_schema_url).group("domain_entity")
        self.module = url_validation_regex.match(metadata_schema_url).group("module")
        self.version = url_validation_regex.match(metadata_schema_url).group("version")
        self.url = metadata_schema_url

    def get_module(self):
        return self.module

    def get_dictionary_representation_of_descriptor(self):
        """ Returns a dictionary representation of the current schema descriptor object. """
        return self.__dict__


class SimplePropertyDescriptor(Descriptor):
    """ A Descriptor encapsulating information about a simple property of a metadata schema. A simple property is
    designated as having no children properties which arises when the property is associated with its own metadata
    schema.
    """

    def __init__(self, json_data):
        """ Initialize the simply property descriptor using the top level fields in given json data. """
        self.value_type = json_data.get("type")
        self.multivalue = False

        if self.value_type == "array":
            self.multivalue = True
            # Get the type of elements in the array which is nested inside the "items" key.
            self.value_type = json_data["items"]["type"]

        self.format = json_data.get("format")
        self.user_friendly = json_data.get("user_friendly")
        self.description = json_data.get("description")
        self.example = json_data.get("example")
        self.guidelines = json_data.get("guidelines")

        # For now, required, external_reference and identifiable are set to false because the value of these properties
        # exist in the parent metadata schema and not in the property description itself. They will be back-populated
        # later.
        self.required = False
        self.identifiable = False
        self.external_reference = False

    def get_dictionary_representation_of_descriptor(self):
        """ Only include information in the class where the value is not None or empty OR if the value is a boolean
        since in that case, both True and False are valid values."""

        return dict((key, value) for (key, value) in self.__dict__.items() if value or isinstance(value, bool))


class ComplexPropertyDescriptor(SimplePropertyDescriptor, Descriptor):
    """ A Descriptor encapsulating information about a complex property of a metadata schema. A complex property
    means that there exists an entire metadata schema to describe the property itself and usually contains children
    properties."""

    def __init__(self, json_data):
        super().__init__(json_data)

        # Populate metadata/information about the schema itself, derived from the URL
        if "$id" in json_data.keys():
            self.schema = SchemaTypeDescriptor(json_data["$id"])
        elif "id" in json_data.keys():
            self.schema = SchemaTypeDescriptor(json_data["id"])
        else:
            self.schema = None

        # Add required fields
        self.required_properties = json_data.get("required")

        # Add children properties
        self.children_properties = {}
        if "properties" in json_data.keys():
            for property_name, property_values in json_data["properties"].items():
                if "$schema" in property_values or "schema" in property_values:
                    child_property_descriptor = ComplexPropertyDescriptor(property_values)
                elif "items" in property_values \
                        and ("$schema" in property_values["items"] or "schema" in property_values["items"]):
                    child_property_descriptor = ComplexPropertyDescriptor(property_values["items"])
                    child_property_descriptor.multivalue = True
                    if child_property_descriptor.user_friendly is None \
                            and "user_friendly" in property_values.keys():
                        child_property_descriptor.user_friendly = property_values["user_friendly"]
                else:
                    child_property_descriptor = SimplePropertyDescriptor(property_values)

                # Make it required if the child property name is in the list of required properties
                if self.required_properties and property_name in self.required_properties:
                    child_property_descriptor.required = True

                # Make the property identifiable if the child property name is one of the listed hardcoded
                # identifiable properties
                if property_name in IDENTIFIABLE_PROPERTIES:
                    child_property_descriptor.identifiable = True

                self.children_properties[property_name] = child_property_descriptor

    def get_schema_module_name(self):
        return self.schema.get_module()

    def get_dictionary_representation_of_descriptor(self):
        """ Returns a representation of the class as a dictionary with the following caveats:
        1) If the value of a key is None or empty but is NOT a boolean, then the attribute it omitted from the
        dictionary.
        2) If the value is of a SchemaTypeDescriptor type, convert it to a dictionary.
        3) Any child descriptors are flattened from being a list to simply added attributes where the key is the
        metadata schema name and the dictionary is the corresponding descriptor.
        """

        dictionary_representation = {}

        for (key, value) in self.__dict__.items():
            if key == "children_properties":
                for child_key, child_value in value.items():
                    self.add_key_value_to_dictionary_if_valid(child_key, child_value, dictionary_representation)
            else:
                self.add_key_value_to_dictionary_if_valid(key, value, dictionary_representation)

        return dictionary_representation

    @staticmethod
    def add_key_value_to_dictionary_if_valid(key, value, dictionary):
        if not value and not isinstance(value, bool):
            return
        if issubclass(type(value), Descriptor):
            dictionary[key] = value.get_dictionary_representation_of_descriptor()
        else:
            dictionary[key] = value



from datetime import datetime, timedelta


class TokenManager:
    def __init__(self, token_client):
        self.token_client = token_client
        self.token = None
        self.TOKEN_DURATION = 3600 * 1000  # 1hr in ms
        self.REFRESH_PERIOD = 60 * 20 * 1000  # 20 min in ms

    def get_token(self):
        token = None
        if self.token and not self.token.is_expired():
            token = self.token
        else:
            token_value = self.token_client.retrieve_token()
            token = self._create_token(token_value)
            self.token = token
        return token.value

    def _create_token(self, value):
        return Token(value=value,
                     token_duration=self.TOKEN_DURATION,
                     refresh_period=self.REFRESH_PERIOD)


class Token:
    def __init__(self, value, token_duration, refresh_period):
        self.value = value
        self.created_at = datetime.now()
        self.token_duration = token_duration
        self.refresh_period = refresh_period

    def is_expired(self):
        now = datetime.now() + timedelta(milliseconds=self.refresh_period)
        expires_at = self.created_at + timedelta(milliseconds=self.token_duration)
        return now > expires_at



import xlsxwriter

from hca_ingest.template.tab_config import TabConfig
from .schema_template import SchemaTemplate

DEFAULT_INGEST_URL = "http://api.ingest.data.humancellatlas.org"
DEFAULT_SCHEMAS_ENDPOINT = "/schemas/search/latestSchemas"
DEFAULT_MIGRATIONS_URL = "https://schema.humancellatlas.org/property_migrations"
MIN_WIDTH_OF_COLUMN = 25


class SpreadsheetBuilder():

    def build(self, schema_template):
        """ Builds an empty spreadsheet. """
        raise NotImplementedError("Subclasses of SpreadsheetBuilder are required to override this method.")

    def create_initial_spreadsheet(self, output_file, hide_row):
        """
        Creates an initial empty spreadsheet with a set formatting.

        :param output_file: A string representing the name of the Excel file that will be generated by this class.
        :param hide_row: A boolean where when true, hides the third row in the generated spreadsheet which contains
                         the fully qualified name of the metadata schema field.
        """

        self.spreadsheet = xlsxwriter.Workbook(output_file)
        self.header_format = self.spreadsheet.add_format(
            {'bold': True, 'bg_color': '#D0D0D0', 'font_size': 12, 'valign': 'vcenter', 'text_wrap': True})
        self.locked_format = self.spreadsheet.add_format({'locked': True})
        self.desc_format = self.spreadsheet.add_format(
            {'font_color': '#808080', 'italic': True, 'text_wrap': True, 'font_size': 12, 'valign': 'top'})
        self.guide_format = self.spreadsheet.add_format(
            {'bold': True, 'bg_color': '#D0D0D0', 'font_size': 12, 'valign': 'vcenter'}
        )
        self.include_schemas_tab = False
        self.hidden_row = hide_row
        self.header_row_height = 50
        self.guide_row_height = 30

    def save_spreadsheet(self):
        self.spreadsheet.close()

    def generate_spreadsheet(self, schema_urls=None, tabs_template=None, include_schemas_tab=False,
                             schema_template=None):
        """
        Given a template that represents the tabs configuration in the desired spreadsheet and a metadata schema,
        generates the respective spreadsheet. If include_schema_tab is set to True, the spreadsheet will have an
        additional sheet containing the URLs of the metadata schemas that were used to generate the spreadsheet.

        :param schema_urls: A list of strings where each string represents a URL containing a JSON-formatted metadata
                            schema.
        :param tabs_template: A string representing a YAML file that contains a configuration specifying how the tabs in
                              the generated spreadsheet should look and what information/columns it should contain.
        :param include_schemas_tab: A boolean, when true, includes an additional sheet at the end of the spreadsheet
                                    that contains a list of metadata schemas that were used to create the spreadsheet.
        :param schema_template: A SchemaTemplate object that has already been created that can directly be used to
                                generate the spreadsheet.
        """

        self.include_schemas_tab = include_schemas_tab
        if tabs_template and schema_urls:
            tabs_parser = TabConfig()
            tabs = tabs_parser.load(tabs_template)
            template = SchemaTemplate(metadata_schema_urls=schema_urls, tab_config=tabs)
        elif schema_urls:
            template = SchemaTemplate(metadata_schema_urls=schema_urls, migrations_url=DEFAULT_MIGRATIONS_URL)
        elif schema_template:
            template = schema_template
        else:
            template = SchemaTemplate()

        self.build(template)

    def get_value_for_column(self, template, column_name, property):
        """ Lookup the value of a column name and property in a SchemaTemplate. `lookup` throws an exception if no
        property exists so capture the exception and return an empty string if no property exists for the given column
        and template.
        :param template: a schema template object
        :param column_name: the full programmatic name of the column
        :param property: the property you want to retrieve eg. example, description
        """
        try:
            if ".text" in column_name:
                value = template.lookup_property_attributes_in_metadata(column_name.replace('.text', '') +
                                                                        "." + property)
                return str(value) if value else ""
            else:
                value = template.lookup_property_attributes_in_metadata(column_name + "." + property)
            return str(value) if value else ""
        except Exception:
            try:
                value = template.lookup_property_attributes_in_metadata(
                    column_name.replace('.text', '') + "." + property)
                return str(value) if value else ""
            except Exception:
                print("No property " + property + " for " + column_name)
                return ""

    def get_user_friendly_column_name(self, template, column_name, primary_schema=None):
        # TODO(maniarathi): Make this description better.
        """ Given a column name derived originally from a metadata schema file that will be inputted as a column name
        into the generated spreadsheet, turn it into a user friendly name. """

        if '.text' in column_name:
            key = column_name.replace('.text', '') + ".user_friendly"
        elif '.ontology_label' in column_name:
            key = column_name.replace('.ontology_label', '') + ".user_friendly"
        elif '.ontology' in column_name:
            key = column_name.replace('.ontology', '') + ".user_friendly"
        else:
            key = column_name + ".user_friendly"

        try:
            try:
                uf = template.lookup_property_attributes_in_metadata(key)
            except Exception:
                uf = template.lookup_property_attributes_in_metadata(column_name + ".user_friendly")

            wrapper = ".".join(column_name.split(".")[:-1])
            if template.lookup_property_attributes_in_metadata(wrapper)['schema']['module'] == 'purchased_reagents' \
                    and not template.lookup_property_attributes_in_metadata(wrapper)['multivalue']:
                uf = template.lookup_property_attributes_in_metadata(wrapper)['user_friendly'] + " - " + uf

            if template.lookup_property_attributes_in_metadata(wrapper)['schema']['module'] == 'barcode'\
                    and not template.lookup_property_attributes_in_metadata(wrapper)['multivalue']:
                uf = template.lookup_property_attributes_in_metadata(wrapper)['user_friendly'] + " - " + uf

            # Add exception for modules that add modules e.g. "cell_suspension.cell_morphology.cell_size_unit.text"
            if len(column_name.split(".")) == 4:
                split_column_name = column_name.split(".")
                module_path = ".".join(split_column_name[0:2])
                module_schema_url = template.lookup_property_attributes_in_metadata(module_path)['schema']['url']
                module_template = SchemaTemplate(metadata_schema_urls=[module_schema_url])
                uf = module_template.meta_data_properties[split_column_name[1]][split_column_name[2]]['user_friendly']

            if '.ontology_label' in column_name and 'ontology label' not in uf:
                uf = uf + " ontology label"
            elif '.ontology' in column_name and 'ontology' not in uf:
                uf = uf + " ontology ID"

            if "Biomaterial " in uf:
                schema_name = column_name.split(".")[0]

                for schema in template.tabs:
                    if schema_name == list(schema.keys())[0]:
                        schema_uf = schema[schema_name]['display_name']
                uf = uf.replace("Biomaterial", schema_uf)

                if primary_schema != schema_name:
                    uf = "Input " + uf

            if "Protocol " in uf:
                schema_name = column_name.split(".")[0]

                for schema in template.tabs:
                    if schema_name == list(schema.keys())[0]:
                        schema_uf = schema[schema_name]['display_name']
                uf = uf.replace("Protocol", schema_uf)

            return uf
        except Exception as e:
            print(e)
            try:
                return uf
            except:
                return column_name

    def generate_and_add_schema_worksheet_to_spreadsheet(self, schema_urls):
        worksheet = self.spreadsheet.add_worksheet("Schemas")
        worksheet.write(0, 0, "Schemas")
        for index, url in enumerate(schema_urls):
            worksheet.write(index + 1, 0, url)



from typing import List

from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet

from hca_ingest.importer.spreadsheet.ingest_workbook import SCHEMAS_WORKSHEET
from hca_ingest.importer.spreadsheet.ingest_worksheet import START_DATA_ROW
from .entity import Entity
from .flattener import Flattener

HEADER_ROW_NO = 4


class XlsDownloader:
    def __init__(self):
        self.flattener = Flattener()

    def convert_json(self, metadata_list: List[Entity]):
        return self.flattener.flatten(metadata_list)

    @staticmethod
    def create_workbook(input_json: dict) -> Workbook:
        workbook = Workbook()
        workbook.remove(workbook.active)

        for ws_title, ws_elements in input_json.items():
            if ws_title == 'Project':
                worksheet: Worksheet = workbook.create_sheet(title=ws_title, index=0)
            elif ws_title == SCHEMAS_WORKSHEET:
                continue
            else:
                worksheet: Worksheet = workbook.create_sheet(title=ws_title)
            XlsDownloader.add_worksheet_content(worksheet, ws_elements)

        XlsDownloader.generate_schemas_worksheet(input_json, workbook)
        return workbook

    @staticmethod
    def generate_schemas_worksheet(input_json: dict, workbook: Workbook):
        schemas = input_json.get(SCHEMAS_WORKSHEET)
        if not schemas:
            raise ValueError('The schema urls are missing')
        schemas_worksheet = workbook.create_sheet(SCHEMAS_WORKSHEET)
        schemas_worksheet.cell(row=1, column=1, value=SCHEMAS_WORKSHEET)
        for row_num, schema in enumerate(schemas, start=2):
            schemas_worksheet.cell(row=row_num, column=1, value=schema)

    @staticmethod
    def add_worksheet_content(worksheet: Worksheet, ws_elements: dict):
        headers = ws_elements.get('headers')
        XlsDownloader.__add_header_row(worksheet, headers)
        all_values = ws_elements.get('values')

        for row_number, row_values in enumerate(all_values, start=START_DATA_ROW):
            XlsDownloader.__add_row_content(worksheet, headers, row_number, row_values)

    @staticmethod
    def __add_header_row(worksheet, headers: list):
        for col, header in enumerate(headers, start=1):
            worksheet.cell(row=HEADER_ROW_NO, column=col, value=header)

    @staticmethod
    def __add_row_content(worksheet, headers: list, row_number: int, values: dict):
        for header, value in values.items():
            index = headers.index(header)
            worksheet.cell(row=row_number, column=index + 1, value=value)



import logging
from typing import List

from hca_ingest.api.ingestapi import IngestApi
from hca_ingest.importer.submission.entity import Entity
from hca_ingest.importer.submission.entity_map import EntityMap

format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(format=format)


class Submission(object):
    def __init__(self, ingest_api: IngestApi, submission_url: str):
        self.ingest_api = ingest_api
        self.submission_url = submission_url
        self.metadata_dict = {}
        self.manifest = None
        self.logger = logging.getLogger(__name__)

    def is_update(self):
        submission = self.ingest_api.get_submission(self.submission_url)
        return submission.get('isUpdate', False)

    def get_submission_url(self):
        return self.submission_url

    def add_entity(self, entity: Entity):
        self.metadata_dict[entity.type + '.' + entity.id] = entity
        return entity

    def get_entity(self, entity_type: str, id: str):
        key = entity_type + '.' + id
        return self.metadata_dict[key]

    def define_manifest(self, entity_map: EntityMap):
        # TODO provide a better way to serialize
        manifest_json = {
            'totalCount': entity_map.count_total(),
            'expectedBiomaterials': entity_map.count_entities_of_type('biomaterial'),
            'expectedProcesses': entity_map.count_entities_of_type('process'),
            'expectedFiles': entity_map.count_entities_of_type('file'),
            'expectedProtocols': entity_map.count_entities_of_type('protocol'),
            'expectedProjects': entity_map.count_entities_of_type('project'),
            'expectedLinks': entity_map.count_links(),
            'actualLinks': 0
        }

        self.manifest = self.ingest_api.create_submission_manifest(self.submission_url, manifest_json)
        return self.manifest

    def get_entities(self) -> List[Entity]:
        return self.metadata_dict.values()



import os
from functools import reduce

import jsonschema
import requests

from .errorreport import ErrorReport
from .validationreport import ValidationReport

BUNDLE_SCHEMA_BASE_URL = "https://schema.humancellatlas.org/bundle/%s/"
BUNDLE_SCHEMA_VERSION = "5.1.0"
BUNDLE_SCHEMA_VERSION = os.environ.get('BUNDLE_SCHEMA_VERSION', BUNDLE_SCHEMA_VERSION)
BUNDLE_SCHEMA_BASE_URL = os.environ.get('BUNDLE_SCHEMA_BASE_URL', BUNDLE_SCHEMA_BASE_URL % BUNDLE_SCHEMA_VERSION)


class BundleValidator:

    def validate(self, metadata, schema_type, version=BUNDLE_SCHEMA_VERSION):

        """
        given a json document(metadata) and a json-schema(schema), validates the
        schema and returns a ValidationReport
        """
        schema = self.load_bundle_schema(schema_type, version)

        validator = jsonschema.Draft4Validator(schema=schema)
        if validator.is_valid(instance=metadata):
            ValidationReport.validation_report_ok()
            return True
        else:
            validation_report = ValidationReport()
            validation_report.validation_state = "INVALID"

            for error in validator.iter_errors(instance=metadata):
                validation_report.error_reports.append(
                    ErrorReport(self.generate_error_message(error), error,
                                "schema validation"))

            return validation_report
            # return False

    def generate_error_message(self, error):
        """
        Given an error object, generates an error message
        :param error: a jsonschema ValidationError
        :return: error message string generated from the error
        """
        path_to_error_in_document = reduce((lambda key1, key2: str(key1) + "." + str(key2)),
                                           error.absolute_path) if len(error.absolute_path) > 0 else "root of document"
        return "Error: " + error.message + " at " + path_to_error_in_document

    def extract_schema_url_from_document(self, metadata_document):
        try:
            return metadata_document["core"]["schema_url"]
        except KeyError:
            raise ("Could not find schema_url")

    def get_schema_from_url(self, schema_url):
        return requests.get(schema_url).json()

    def load_bundle_schema(self, schema_type, version):

        url = BUNDLE_SCHEMA_BASE_URL + schema_type
        if version:
            url = "https://schema.humancellatlas.org/bundle/" + version + "/" + schema_type
        schema = self.get_schema_from_url(url)

        return schema



# TODO Check how we can refactor to merge MetadataEntity and Entity
class Entity(object):
    def __init__(self, entity_type, entity_id, content, ingest_json=None, links_by_entity=None,
                 direct_links=None, external_links=None, is_reference=False, is_linking_reference=False, linking_details=None, concrete_type=None,
                 spreadsheet_location=None):
        self.type = entity_type
        self.id = entity_id
        self.content = content
        self._prepare_links_by_entity(links_by_entity)
        self._prepare_direct_links(direct_links)
        self._prepare_linking_details(linking_details)
        self._prepare_external_links(external_links)
        self.ingest_json = ingest_json
        self.is_reference = is_reference  # if entity is in a row and has uuid
        self.is_linking_reference = is_linking_reference  # if the entity uuid is only specified in a linking column
        self.concrete_type = concrete_type
        self.spreadsheet_location = spreadsheet_location

    def _prepare_links_by_entity(self, links_by_entity):
        self.links_by_entity = {}
        if links_by_entity is not None:
            self.links_by_entity.update(links_by_entity)

    def _prepare_direct_links(self, direct_links):
        self.direct_links = []
        if direct_links is not None:
            self.direct_links.extend(direct_links)

    def _prepare_external_links(self, external_links):
        self.external_links = {}
        if external_links is not None:
            self.external_links.update(external_links)

    def _prepare_linking_details(self, linking_details):
        self.linking_details = {}
        if linking_details is not None:
            self.linking_details.update(linking_details)

    @property
    def uuid(self):
        return self.ingest_json.get('uuid', {}).get('uuid')

    @property
    def url(self):
        return self.ingest_json['_links']['self']['href']

    @property
    def is_new(self):
        return not self.is_reference and not self.is_linking_reference

    def add_link(self, type:str, entity_id:str):
        links = self.links_by_entity.get(type, [])
        links.append(entity_id)
        self.links_by_entity[type] = links


#!/usr/bin/env python
"""
Description goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "04/05/2018"


def put(d, keys, item):
    if "." in keys:
        key, rest = keys.split(".", 1)
        if key not in d:
            d[key] = {}
        put(d[key], rest, item)
    else:
        d[keys] = item


def get(d, keys):
    if "." in keys:
        key, rest = keys.split(".", 1)
        return get(d[key], rest)
    else:
        return d[keys]


class DotDict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



import string

from openpyxl import Workbook

HEADER = list(string.ascii_uppercase)


def create_worksheet(name, rows, start_row=1):
    rows = rows or []
    start_row = start_row or 1

    workbook = Workbook()
    worksheet = workbook.create_sheet(name)

    for row_idx, row in enumerate(rows):
        for col_idx, cell_value in enumerate(row):
            header = HEADER[col_idx]
            worksheet[f'{header}{start_row + row_idx}'] = cell_value

    return worksheet



#!/usr/bin/env python
"""
Description goes here
"""
from requests import HTTPError

__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "12/09/2017"

import json
import logging
import os
from time import time
from urllib.parse import urljoin

import requests
import urllib3.util.retry as retry

DEFAULT_STAGING_URL = os.environ.get('STAGING_API', 'https://upload.dev.data.humancellatlas.org')
DEFAULT_STAGING_VERSION = os.environ.get('STAGING_API_VERSION', 'v1')
INGEST_API_KEY = os.environ.get('INGEST_API_KEY', 'zero-pupil-until-funny')


class RetryPolicy(retry.Retry):
    def __init__(self, retry_after_status_codes={301}, **kwargs):
        super(RetryPolicy, self).__init__(**kwargs)
        self.RETRY_AFTER_STATUS_CODES = frozenset(retry_after_status_codes | retry.Retry.RETRY_AFTER_STATUS_CODES)


class UploadResponseParseException(Exception):
    pass


class FileDescription:

    def __init__(self, checksums: dict, content_type: str, name: str, size, url: str):
        self.checksums = checksums
        self.content_type = content_type
        self.name = name
        self.size = size
        self.url = url

    @staticmethod
    def parse_upload_response(res: dict):
        try:
            return FileDescription(res['checksums'], res['content_type'], res['name'], res['size'], res['url'])
        except (KeyError, TypeError) as e:
            raise UploadResponseParseException() from e


class MetadataFileStagingRequest:
    def __init__(self, staging_area_uuid, filename, metadata_json, metadata_type):
        self.staging_area_uuid = staging_area_uuid
        self.filename = filename
        self.metadata_json = metadata_json
        self.metadata_type = metadata_type


class FileUploadFailed(Exception):
    pass


class StagingApi:
    def __init__(self, url=None, apikey=None, apiversion=None):
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        logging.basicConfig(formatter=formatter)

        retry_policy = RetryPolicy(
            total=100,  # seems that this has a default value of 10,
            # setting this to a very high number so that it'll respect the status retry count
            status=17,  # status is the no. of retries if response is in status_forcelist,
            # this count will retry for ~20mins with back off timeout within
            read=10,
            status_forcelist=[500, 502, 503, 504],
            backoff_factor=0.6,
            method_whitelist=frozenset(['HEAD', 'GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'TRACE'])
        )

        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=retry_policy)
        self.session.mount('https://', adapter)

        self.logger = logging.getLogger(__name__)

        if not apiversion and DEFAULT_STAGING_VERSION:
            self.apiversion = DEFAULT_STAGING_VERSION
        self.apiversion = apiversion if apiversion else "v1"

        if not url and DEFAULT_STAGING_URL:
            url = DEFAULT_STAGING_URL
            # expand interpolated env vars
            self.url = os.path.expandvars(url)
            self.logger.info(f'Using {url} for staging API')
        self.url = url if url else 'https://upload.dev.data.humancellatlas.org'

        if not apikey and INGEST_API_KEY:
            apikey = INGEST_API_KEY
        self.apikey = apikey if apikey else 'zero-pupil-until-funny'

        self.header = {'Api-Key': self.apikey, 'Content-type': 'application/json'}

    def createStagingArea(self, submissionId):
        start_time = time()
        self.logger.info('Creating staging area!')
        base = urljoin(self.url, self.apiversion + '/area/' + submissionId)

        r = self.session.post(base, headers=self.header)
        r.raise_for_status()
        self.logger.info(f'Staging area created!: {base}')
        self.logger.info("Execution Time: %s seconds" % (time() - start_time))
        return r.json()

    def deleteStagingArea(self, submissionId):
        self.logger.info('Deleting staging area!')
        base = urljoin(self.url, self.apiversion + '/area/' + submissionId)
        r = self.session.delete(base, headers=self.header)
        r.raise_for_status()
        self.logger.info('Staging area deleted!')
        return base

    def stageFileRequest(self, file_stage_request: 'MetadataFileStagingRequest'):
        return self.stageFile(file_stage_request.staging_area_uuid,
                              file_stage_request.filename,
                              file_stage_request.metadata_json,
                              file_stage_request.metadata_type)

    def stageFile(self, staging_area_uuid, file_name, body, type):
        file_url = urljoin(self.url, self.apiversion + '/area/' + staging_area_uuid + "/" + file_name)

        self.logger.info(f'Staging file: {file_url}')

        header = dict(self.header)
        header['Content-type'] = f'application/json; dcp-type="{type}"'

        r = self.session.put(file_url, data=json.dumps(body, indent=4), headers=header)

        try:
            r.raise_for_status()
            res = r.json()
            return FileDescription.parse_upload_response(res)
        except (HTTPError, UploadResponseParseException) as exception:
            raise FileUploadFailed(f'File upload failed for {file_url}.') from exception

    def hasStagingArea(self, submissionId) -> bool:
        base = urljoin(self.url, self.apiversion + '/area/' + submissionId)
        r = self.session.head(base, headers=self.header)
        return r.status_code == requests.codes.ok



from typing import List

from hca_ingest.importer.conversion.template_manager import TemplateManager
from hca_ingest.importer.submission.entity import Entity
from hca_ingest.importer.submission.entity_map import EntityMap
from hca_ingest.importer.submission.errors import LinkedEntityNotFound, InvalidLinkInSpreadsheet, MultipleProcessesFound

"""
EntityLinker is responsible for converting linking info from spreadsheet to linking in Ingest Core
"""


class EntityLinker(object):

    def __init__(self, template_manager: TemplateManager, entity_map: EntityMap):
        self.template_manager = template_manager
        self.process_id_ctr = 0
        self.entity_map = entity_map

    def convert_spreadsheet_links_to_ingest_links(self):
        for entity in self.entity_map.get_entities():
            self._load_external_links_to_entity_map(entity)
            self._validate_entity_links(entity)
            self._generate_direct_links(entity)
        return self.entity_map

    def _load_external_links_to_entity_map(self, entity: Entity):
        external_links = entity.external_links
        for external_link_type, external_link_uuids in external_links.items():
            for entity_uuid in external_link_uuids:
                external_link_entity = Entity(entity_type=external_link_type,
                                              entity_id=entity_uuid,
                                              content=None,
                                              spreadsheet_location=entity.spreadsheet_location,
                                              is_linking_reference=True)
                external_link_entity.add_link(external_link_type, entity_uuid)
                self.entity_map.add_entity(external_link_entity)

    # direct_links maps to ingest db model
    def _generate_direct_links(self, entity: Entity):
        project = self.entity_map.get_project()

        self._link_entity_to_project(entity, project)
        self._link_supplementary_file_to_project(entity, project)

        external_links = entity.external_links or {}
        links_by_entity = entity.links_by_entity or {}

        input_biomaterial_ids = []
        input_biomaterial_ids.extend(external_links.get('biomaterial', []))
        input_biomaterial_ids.extend(links_by_entity.get('biomaterial', []))

        input_file_ids = []
        input_file_ids.extend(external_links.get('file', []))
        input_file_ids.extend(links_by_entity.get('file', []))

        if input_biomaterial_ids or input_file_ids:
            process = self._create_or_get_process(entity)
            self.entity_map.add_entity(process)
            self._link_process_to_project(process, project)
            self._link_entity_as_output_to_process(entity, process)
            self._link_protocols_to_process(entity, process)
            self._link_input_biomaterials_to_entity(input_biomaterial_ids, process)
            self._link_input_files_to_entity(input_file_ids, process)

    def _link_entity_as_output_to_process(self, entity: Entity, linking_process: Entity):
        entity.direct_links.append({
            'entity': linking_process.type,
            'id': linking_process.id,
            'relationship': 'derivedByProcesses'
        })

    def _link_input_files_to_entity(self, linked_file_ids: List[str], linking_process: Entity):
        for linked_file_id in linked_file_ids:
            linked_file_entity = self.entity_map.get_entity('file', linked_file_id)
            linked_file_entity.direct_links.append({
                'entity': linking_process.type,
                'id': linking_process.id,
                'relationship': 'inputToProcesses'
            })

    def _link_input_biomaterials_to_entity(self, linked_biomaterial_ids: List[str], linking_process: Entity):
        for linked_biomaterial_id in linked_biomaterial_ids:
            linked_biomaterial_entity = self.entity_map.get_entity('biomaterial', linked_biomaterial_id)
            linked_biomaterial_entity.direct_links.append({
                'entity': linking_process.type,
                'id': linking_process.id,
                'relationship': 'inputToProcesses'
            })

    def _link_protocols_to_process(self, entity: Entity, process: Entity):
        links_by_entity = entity.links_by_entity
        external_links = entity.external_links

        protocol_ids = []
        protocol_ids.extend(external_links.get('protocol', []))
        protocol_ids.extend(links_by_entity.get('protocol', []))

        for protocol_id in protocol_ids:
            process.direct_links.append({
                'entity': 'protocol',
                'id': protocol_id,
                'relationship': 'protocols'
            })

    def _link_process_to_project(self, linking_process: Entity, project: Entity):
        linking_process.direct_links.append({
            'entity': 'project',
            'id': project.id,
            'relationship': 'project',
            'is_collection': False
        })
        # TODO: Remove when process.projects is deprecated
        linking_process.direct_links.append({
            'entity': 'project',
            'id': project.id,
            'relationship': 'projects'
        })

    def _link_supplementary_file_to_project(self, entity: Entity, project: Entity):
        if project and entity.concrete_type == 'supplementary_file':
            project.direct_links.append({
                'entity': 'file',
                'id': entity.id,
                'relationship': 'supplementaryFiles'
            })

    def _link_entity_to_project(self, entity: Entity, project: Entity):
        if project and entity.type != 'project':
            entity.direct_links.append({
                'entity': 'project',
                'id': project.id,
                'relationship': 'project',
                'is_collection': False
            })
            # TODO: Remove when biomaterial/process.projects is deprecated
            # https://github.com/ebi-ait/dcp-ingest-central/issues/88
            if entity.type == 'biomaterial' or entity.type == 'process':
                entity.direct_links.append({
                    'entity': 'project',
                    'id': project.id,
                    'relationship': 'projects'
                })

    def _validate_entity_links(self, entity: Entity):
        if entity.is_reference:
            return

        links_by_entity = entity.links_by_entity

        for link_entity_type, link_entity_ids in links_by_entity.items():
            for link_entity_id in link_entity_ids:
                if not link_entity_type == 'process':  # it is expected that no processes are defined in any tab,
                    # these will be created later
                    if not self._is_valid_spreadsheet_link(entity.type, link_entity_type):
                        raise InvalidLinkInSpreadsheet(entity, link_entity_type, link_entity_id)
                    if not self.entity_map.get_entity(link_entity_type, link_entity_id):
                        raise LinkedEntityNotFound(entity, link_entity_type, link_entity_id)
                    if not self.entity_map.get_entity(link_entity_type, link_entity_id):
                        raise LinkedEntityNotFound(entity, link_entity_type, link_entity_id)

                if link_entity_type == 'process' and not len(link_entity_ids) == 1:
                    raise MultipleProcessesFound(entity, link_entity_ids)

    def _create_or_get_process(self, entity: Entity) -> Entity:
        external_links = entity.external_links
        processes = external_links.get('process', [])
        external_process_id = processes[0] if len(processes) > 0 else None

        links_by_entity = entity.links_by_entity
        process_id = links_by_entity['process'][0] if links_by_entity.get('process') else None

        if not process_id:
            process_id = self._generate_empty_process_id()

        linking_details = entity.linking_details

        process = self.entity_map.get_entity('process', process_id)

        if not process:
            process_json = self._create_process_json(process_id, linking_details)
            process = Entity(
                entity_type='process',
                entity_id=external_process_id or process_id,
                content=process_json,
                is_linking_reference=bool(external_process_id),
                is_reference=bool(external_process_id)
            )

        return process

    @staticmethod
    def _is_valid_spreadsheet_link(from_entity_type, to_entity_type):
        VALID_ENTITY_LINKS_MAP = [
            'biomaterial-biomaterial',
            'file-biomaterial',
            'file-file',
            'biomaterial-process',
            'biomaterial-protocol',
            'file-process',
            'file-protocol',
        ]
        link_key = from_entity_type + '-' + to_entity_type

        return link_key in VALID_ENTITY_LINKS_MAP

    def _create_process_json(self, process_id: str, linking_details: dict):
        schema_type = 'process'
        described_by = self.template_manager.get_schema_url(schema_type)

        if linking_details:
            if not linking_details.get('process_core'):
                linking_details['process_core'] = {}

            linking_details['process_core']['process_id'] = process_id
            linking_details['schema_type'] = schema_type
            linking_details['describedBy'] = described_by
        else:
            process_core = {'process_id': process_id}
            linking_details = {
                "process_core": process_core,
                "schema_type": schema_type,
                "describedBy": described_by
            }

        return linking_details

    def _generate_empty_process_id(self):
        self.process_id_ctr += 1
        return 'process_id_' + str(self.process_id_ctr)



from typing import List


class Entity:
    def __init__(self, content: dict, uuid: str, id: str = None):
        self._content = content
        self._uuid = uuid
        self._id = id
        self._input_biomaterials = None
        self._input_files = None
        self._process = None
        self._protocols = None

    @classmethod
    def from_json(cls, entity_json: dict):
        content = entity_json.get('content')
        uuid = entity_json.get('uuid', {})
        links = entity_json.get('_links', {})
        self_link = links.get('self', {})
        self_href = self_link.get('href')
        id = self_href.split('/')[-1] if self_href else None;
        return cls(content, uuid.get('uuid'), id)

    @classmethod
    def from_json_list(cls, entity_json_list: List[dict]):
        return [Entity.from_json(e) for e in entity_json_list]

    @property
    def content(self):
        return self._content

    @property
    def uuid(self):
        return self._uuid

    @property
    def id(self):
        return self._id

    @property
    def input_biomaterials(self):
        return self._input_biomaterials

    @property
    def input_files(self):
        return self._input_files

    @property
    def protocols(self):
        return self._protocols

    @property
    def process(self):
        return self._process

    @property
    def concrete_type(self):
        return self._content.get('describedBy').rsplit('/', 1)[-1]

    @property
    def domain_type(self):
        return self._content.get('describedBy').split('/')[4]

    def set_input(self, input_biomaterials, input_files, process, protocols):
        assert isinstance(process, Entity)
        assert all(isinstance(protocol, Entity) for protocol in protocols)
        assert all(isinstance(input_biomaterial, Entity) for input_biomaterial in input_biomaterials)
        assert all(isinstance(input_file, Entity) for input_file in input_files)
        self._input_files = input_files
        self._input_biomaterials = input_biomaterials
        self._process = process
        self._protocols = protocols



import copy

from hca_ingest.importer.data_node import DataNode
from hca_ingest.importer.spreadsheet.ingest_worksheet import IngestRow

TYPE_UNDEFINED = 'undefined'


class MetadataEntity:

    # TODO enforce definition of concrete and domain types for all MetadataEntity
    # It's only currently done this way to minimise friction with other parts of the system
    def __init__(self, concrete_type=TYPE_UNDEFINED, domain_type=TYPE_UNDEFINED, object_id=None,
                 content={}, links={}, external_links={}, linking_details={}, row: IngestRow = None,
                 is_reference=False, is_linking_reference=False, is_module=False):
        self._concrete_type = concrete_type
        self._domain_type = domain_type
        self.object_id = object_id
        self._content = DataNode(defaults=copy.deepcopy(content))
        self._links = copy.deepcopy(links)
        self._external_links = copy.deepcopy(external_links)
        self._linking_details = DataNode(defaults=copy.deepcopy(linking_details))
        self._spreadsheet_location = {
            'row_index': row.index,
            'worksheet_title': row.worksheet_title,
        } if row else None
        self.is_reference = is_reference
        self.is_linking_reference = is_linking_reference
        self._is_module = is_module

    @property
    def concrete_type(self):
        return self._concrete_type

    @property
    def domain_type(self):
        return self._domain_type

    @property
    def content(self):
        return copy.deepcopy(self._content)

    def get_content(self, content_property):
        return self._content[content_property]

    def define_content(self, content_property, value):
        self._content[content_property] = value

    def define_linking_detail(self, link_property, value):
        self._linking_details[link_property] = value

    @property
    def linking_details(self):
        return self._linking_details.as_dict()

    def get_linking_detail(self, link_property):
        return self._linking_details[link_property]

    @property
    def links(self):
        return copy.deepcopy(self._links)

    def get_links(self, link_entity_type):
        return self._links.get(link_entity_type)

    def add_links(self, link_entity_type, new_links):
        self._do_add_links(self._links, link_entity_type, new_links)

    @property
    def is_module(self):
        return self._is_module

    @property
    def external_links(self):
        return copy.deepcopy(self._external_links)

    def get_external_links(self, link_entity_type):
        return self._external_links.get(link_entity_type)

    def add_external_links(self, link_entity_type, new_links):
        self._do_add_links(self._external_links, link_entity_type, new_links)

    @staticmethod
    def _do_add_links(link_map, link_entity_type, new_links):
        existent_links = link_map.get(link_entity_type)
        if existent_links is None:
            existent_links = []
            link_map[link_entity_type] = existent_links
        existent_links.extend(new_links)

    def retain_fields(self, *allowed_fields):
        removed_fields = []
        for key in self._content.keys():
            if key not in allowed_fields:
                removed_fields.append({'key': key, 'value': self._content[key]})
                self._content.remove_field(key)
        return removed_fields

    def add_module_entity(self, module_entity):
        for field, value in module_entity.content.as_dict().items():
            module_list = self._content[field]
            if not module_list:
                module_list = []
                self._content[field] = module_list
            module_list.append(value)

    def get_spreadsheet_location(self):
        return self._spreadsheet_location

    def map_for_submission(self):
        return {
            'is_reference': self.is_reference,
            'is_linking_reference': self.is_linking_reference,
            'concrete_type': self.concrete_type,
            'content': self._content.as_dict(),
            'links_by_entity': self.links,
            'external_links_by_entity': self.external_links,
            'linking_details': self.linking_details,
            'spreadsheet_location': self._spreadsheet_location
        }



from openpyxl.workbook import Workbook

from api.ingestapi import IngestApi
from .data_collector import DataCollector
from .downloader import XlsDownloader


class WorkbookDownloader:
    def __init__(self, api: IngestApi):
        self.collector = DataCollector(api)
        self.downloader = XlsDownloader()

    def get_workbook_from_submission(self, submission_uuid: str) -> Workbook:
        entity_dict = self.collector.collect_data_by_submission_uuid(submission_uuid)
        entity_list = list(entity_dict.values())
        flattened_json = self.downloader.convert_json(entity_list)
        return self.downloader.create_workbook(flattened_json)



from typing import List


from hca_ingest.importer.submission.entity import Entity


class EntityMap(object):

    def __init__(self, *entities):
        self.entities_dict_by_type = {}
        if entities is None:
            return
        for entity in entities:
            self.add_entity(entity)

    @staticmethod
    def load(entity_json):
        entity_map = EntityMap()

        for entity_type, entities_dict in entity_json.items():
            for entity_id, entity_body in entities_dict.items():
                entity = Entity(entity_type=entity_type,
                                entity_id=entity_id,
                                content=entity_body.get('content'),
                                external_links=entity_body.get('external_links_by_entity'),
                                links_by_entity=entity_body.get('links_by_entity', {}),
                                is_reference=entity_body.get('is_reference', False),
                                is_linking_reference=entity_body.get('is_linking_reference', False),
                                linking_details=entity_body.get('linking_details', {}),
                                concrete_type=entity_body.get('concrete_type'),
                                spreadsheet_location=entity_body.get(
                                    'spreadsheet_location'))
                entity_map.add_entity(entity)

        return entity_map

    def get_entity_types(self):
        return list(self.entities_dict_by_type.keys())

    def get_entities_of_type(self, type) -> List[Entity]:
        entities = []

        entities_dict = self.entities_dict_by_type.get(type, {})

        for entity_id, entity in entities_dict.items():
            entities.append(entity)

        return entities

    def get_new_entities_of_type(self, type) -> List[Entity]:
        entities = []
        entities_dict = self.entities_dict_by_type.get(type, {})
        for entity_id, entity in entities_dict.items():
            if not (entity.is_reference and entity.is_linking_reference):
                entities.append(entity)

        return entities

    def get_entity(self, type, id) -> Entity:
        if self.entities_dict_by_type.get(type) and self.entities_dict_by_type[type].get(id):
            return self.entities_dict_by_type[type][id]

    def add_entity(self, entity: Entity):
        entities_of_type = self.entities_dict_by_type.get(entity.type)
        if not entities_of_type:
            self.entities_dict_by_type[entity.type] = {}
            entities_of_type = self.entities_dict_by_type.get(entity.type)

        existing_entity = self.get_entity(entity.type, entity.id)
        if existing_entity and existing_entity.is_reference and entity.is_linking_reference:
            existing_entity.is_linking_reference = True
        elif existing_entity and existing_entity.is_linking_reference and entity.is_reference:
            existing_entity.is_reference = entity.is_reference
            existing_entity.content = entity.content
        elif existing_entity:
            entities_of_type[entity.id] = existing_entity
        else:
            entities_of_type[entity.id] = entity

    def get_entities(self) -> List[Entity]:
        all_entities = []
        for entity_type, entities_dict in self.entities_dict_by_type.items():
            all_entities.extend(entities_dict.values())
        return all_entities

    def get_new_entities(self) -> List[Entity]:
        all_entities = []
        for entity_type, entities_dict in self.entities_dict_by_type.items():
            for entity_id, entity in entities_dict.items():
                if entity.is_new:
                    all_entities.append(entity)
        return all_entities

    def get_project(self) -> Entity:
        project_ids = list(self.entities_dict_by_type.get('project', {}).keys())
        return self.get_entity('project', project_ids[0]) if project_ids else None

    def count_total(self) -> int:
        return len(self.get_entities())

    def count_entities_of_type(self, type) -> int:
        return len(self.get_new_entities_of_type(type))

    def count_links(self) -> int:
        count = 0
        for entity in self.get_entities():
            count = count + len(entity.direct_links)
        return count



from datetime import datetime

INGEST_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
INGEST_DATE_FORMAT_SHORT = "%Y-%m-%dT%H:%M:%SZ"

_expected_formats = [INGEST_DATE_FORMAT, INGEST_DATE_FORMAT_SHORT]


def parse_date_string(date_str: str):
    for date_format in _expected_formats:
        try:
            return datetime.strptime(date_str, date_format)
        except ValueError:
            pass
    raise ValueError(f'unknown date format for [{date_str}]')


def date_to_json_string(date: datetime):
    return date.isoformat().replace("+00:00", "Z")




import logging
from typing import Tuple, List

from openpyxl import Workbook
from requests import HTTPError

from hca_ingest.api.ingestapi import IngestApi
from hca_ingest.importer.conversion import template_manager
from hca_ingest.importer.conversion.metadata_entity import MetadataEntity
from hca_ingest.importer.conversion.template_manager import TemplateManager
from hca_ingest.importer.spreadsheet.ingest_workbook import IngestWorkbook
from hca_ingest.importer.spreadsheet.ingest_worksheet import IngestWorksheet
from hca_ingest.importer.submission.entity_linker import EntityLinker
from hca_ingest.importer.submission.entity_map import EntityMap
from hca_ingest.importer.submission.ingest_submitter import IngestSubmitter
from hca_ingest.importer.submission.submission import Submission
from hca_ingest.template.exceptions import UnknownKeySchemaException
from hca_ingest.utils.IngestError import ImporterError, ParserError

format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(format=format)


class XlsImporter:
    """
    XlsImporter is used to convert a contributor's spreadsheet into metadata json entities and to submit those to
    Ingest. Please see https://github.com/HumanCellAtlas/ingest-central/wiki/Data-Contributors-Spreadsheet-Quick-Guide
    for more information on the spreadsheet format.
    """

    def __init__(self, ingest_api: IngestApi):
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)
        self.submitter = IngestSubmitter(self.ingest_api)

    def generate_json(self, file_path, is_update, project_uuid=None, update_project=False):
        ingest_workbook = IngestWorkbook.from_file(file_path)

        try:
            template_mgr = template_manager.build(ingest_workbook.get_schemas(), self.ingest_api)
        except Exception as e:
            raise SchemaRetrievalError(
                f'There was an error retrieving the schema information to process the spreadsheet. {str(e)}')

        workbook_importer = WorkbookImporter(template_mgr)
        spreadsheet_json, errors = workbook_importer.do_import(ingest_workbook, is_update, project_uuid, update_project)

        return spreadsheet_json, template_mgr, errors

    def dry_run_import_file(self, file_path, project_uuid=None):
        spreadsheet_json, template_mgr, errors = self.generate_json(file_path, project_uuid)

        if errors:
            return None, errors

        entity_map = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(template_mgr, entity_map)
        entity_linker.convert_spreadsheet_links_to_ingest_links()

        return entity_map, []

    def import_file(self, file_path, submission_url, is_update=False, project_uuid=None, update_project=False) -> Tuple[
        Submission, TemplateManager]:
        try:
            if project_uuid:
                self.submitter.link_submission_to_project(project_uuid, submission_url)

            submission = None
            template_mgr = None
            spreadsheet_json, template_mgr, errors = self.generate_json(file_path, is_update, project_uuid=project_uuid,
                                                                        update_project=update_project)
            entity_map = EntityMap.load(spreadsheet_json)
            self.ingest_api.delete_submission_errors(submission_url)

            if errors:
                self.report_errors(submission_url, errors)
            elif is_update:
                self.submitter.update_entities(entity_map)
            else:
                entity_linker = EntityLinker(template_mgr, entity_map)
                entity_linker.convert_spreadsheet_links_to_ingest_links()
                submission = self._submit_new_entities(entity_map, submission_url)

            project = entity_map.get_project()
            if project and project_uuid and update_project:
                self.submitter.update_entity(project)

        except HTTPError as httpError:
            self.logger.exception(httpError)
            status = httpError.response.status_code
            text = httpError.response.text
            importer_error = ImporterError(f'Received an HTTP {status} from  {httpError.request.url}: {text}')
            self.ingest_api.create_submission_error(submission_url, importer_error.getJSON())
            return None, template_mgr
        except Exception as e:
            self.ingest_api.create_submission_error(submission_url, ImporterError(str(e)).getJSON())
            self.logger.error(str(e), exc_info=True)
            return None, template_mgr
        finally:
            self.logger.info(f'Submission in {submission_url} is done!')
            return submission, template_mgr

    def _submit_new_entities(self, entity_map, submission_url):
        submission = self.submitter.add_entities(entity_map, submission_url)
        project = entity_map.get_project()
        if project and project.is_new:
            self.submitter.link_submission_to_project(project.uuid, submission_url)

        self.submitter.link_entities(entity_map, submission)
        return submission

    def report_errors(self, submission_url, errors):
        self.logger.info(f'Logged {len(errors)} ParsingErrors.', exc_info=False)
        for error in errors:
            self.ingest_api.create_submission_error(
                submission_url,
                ParserError(error["location"], error["type"], error["detail"]).getJSON()
            )

    @staticmethod
    def update_spreadsheet_with_uuids(submission: Submission, template_mgr: TemplateManager, file_path):
        if not submission:
            return
        wb = IngestWorkbook.from_file(file_path, read_only=False)
        wb.add_entity_uuids(submission)
        wb.add_schemas_worksheet(template_mgr.get_schemas())
        return wb.save(file_path)

    def import_project_from_workbook(self, workbook: Workbook, token: str) -> (str, List[dict]):
        project_metadata_json, errors = self._generate_project_json_from_workbook(workbook)

        if errors:
            return None, errors
        else:
            ingest_project = self.ingest_api.create_project(None, content=project_metadata_json, token=token)
            project_uuid = ingest_project['uuid']['uuid']
            return project_uuid, []

    def _generate_project_json_from_workbook(self, workbook):
        ingest_workbook = IngestWorkbook(workbook)
        template_mgr = self._setup_template_manager_for_project_import()
        workbook_importer = WorkbookImporter(template_mgr)
        spreadsheet_json, errors = workbook_importer.do_import(ingest_workbook, False, worksheet_titles=['Project'])

        if errors:
            return None, errors
        else:
            projects = list(spreadsheet_json.get('project').values())
            project = projects[0] if projects else None
            project_metadata = project.get('content')
            return project_metadata, []

    def _setup_template_manager_for_project_import(self):
        try:
            project_schema_url = self.ingest_api.get_latest_schema_url('type', 'project', 'project')
            template_mgr = template_manager.build([project_schema_url], self.ingest_api)
        except Exception as e:
            raise SchemaRetrievalError(
                f'There was an error retrieving the project schema information to import the project. {str(e)}')
        return template_mgr


_PROJECT_ID = 'project_0'
_PROJECT_TYPE = 'project'


class _ImportRegistry:
    """
    This is a helper class for managing metadata entities during Workbook import.
    """

    def __init__(self, template_mgr: TemplateManager):
        self.template_mgr = template_mgr
        self._submittable_registry = {}
        self._module_registry = {}
        self._module_list = []
        self.project_id = _PROJECT_ID

    def add_submittable(self, metadata: MetadataEntity):
        # TODO no test to check case sensitivity
        domain_type = metadata.domain_type.lower()
        type_map = self._submittable_registry.get(domain_type)
        if not type_map:
            type_map = {}
            self._submittable_registry[domain_type] = type_map
        if domain_type.lower() == _PROJECT_TYPE:
            if not type_map.get(self.project_id):
                metadata.object_id = metadata.object_id or self.project_id
                self.project_id = metadata.object_id
            else:
                raise MultipleProjectsFound()
        type_map[metadata.object_id] = metadata

    def add_submittables(self, metadata_entities):
        for entity in metadata_entities:
            self.add_submittable(entity)

    def add_project_reference(self, project_uuid, project: MetadataEntity = None):
        if project:
            project.object_id = project_uuid
            project.is_linking_reference = True
            project.is_reference = True
        else:
            project = MetadataEntity(domain_type=_PROJECT_TYPE,
                                     concrete_type=_PROJECT_TYPE,
                                     object_id=project_uuid,
                                     is_linking_reference=True)
        self.add_submittable(project)

    def add_module(self, metadata: MetadataEntity):
        if metadata.domain_type.lower() == 'project':
            metadata.object_id = self.project_id
        self._module_list.append(metadata)

    def add_modules(self, module_field_name, metadata_entities):
        all_removed_fields = []
        for entity in metadata_entities:
            removed_fields = entity.retain_fields(module_field_name)
            all_removed_fields.extend(removed_fields)
            self.add_module(entity)
        return all_removed_fields

    def import_modules(self):
        for module_entity in self._module_list:
            type_map = self._submittable_registry.get(module_entity.domain_type, {})
            submittable_entity = type_map.get(module_entity.object_id)
            if submittable_entity:
                submittable_entity.add_module_entity(module_entity)
            else:
                raise LinkToConcreteEntityNotFound(module_entity)

    def flatten(self):
        flat_map = {}
        for domain_type, type_map in self._submittable_registry.items():
            flat_type_map = {object_id: metadata.map_for_submission()
                             for object_id, metadata in type_map.items()}
            flat_map[domain_type] = flat_type_map
        return flat_map

    def has_project(self):
        project_registry = self._submittable_registry.get(_PROJECT_TYPE)
        return project_registry and project_registry.get(self.project_id)


class WorkbookImporter:
    def __init__(self, template_mgr):
        self.worksheet_importer = WorksheetImporter(template_mgr)
        self.template_mgr = template_mgr
        self.logger = logging.getLogger(__name__)

    def do_import(self, workbook: IngestWorkbook, is_update, project_uuid=None, update_project=False,
                  worksheet_titles: List[str] = None):
        registry = _ImportRegistry(self.template_mgr)

        if worksheet_titles:
            importable_worksheets = workbook.select_importable_worksheets(worksheet_titles)
        else:
            importable_worksheets = workbook.importable_worksheets()

        workbook_errors = self.validate_worksheets(is_update, importable_worksheets)

        importable_worksheets = [ws for ws in importable_worksheets]

        for worksheet in importable_worksheets:
            try:
                self._import_worksheet(project_uuid, registry, update_project, workbook_errors, worksheet)
            except Exception as e:
                workbook_errors.append(
                    {"location": f'sheet={worksheet.title}', "type": e.__class__.__name__, "detail": str(e)})

        if not registry.has_project() and not project_uuid:
            e = NoProjectFound()
            workbook_errors.append({"location": "File", "type": e.__class__.__name__, "detail": str(e)})

        if project_uuid and update_project and not registry.has_project():
            workbook_errors.append({"location": "File", "type": "NoProjectWorksheet",
                                    "detail": "The option to update the project was specified but there is no project "
                                              "worksheet found."})

        if project_uuid and not registry.has_project():
            registry.add_project_reference(project_uuid)

        self._import_modules(registry, workbook_errors)

        return registry.flatten(), workbook_errors

    def _import_worksheet(self, project_uuid, registry, update_project, workbook_errors, worksheet):
        self.sheet_in_schemas(worksheet)
        metadata_entities, worksheet_errors = self.worksheet_importer.do_import(worksheet)
        module_field_name = worksheet.get_module_field_name()
        workbook_errors.extend(worksheet_errors)

        if project_uuid and worksheet.is_project_module() and update_project:
            self._register_module(metadata_entities, module_field_name, registry, workbook_errors, worksheet)

        elif project_uuid and worksheet.is_project() and metadata_entities:
            if len(metadata_entities) > 1:
                raise MultipleProjectsFound()

            registry.add_project_reference(project_uuid, metadata_entities[0])

        elif worksheet.is_module_tab():
            self._register_module(metadata_entities, module_field_name, registry, workbook_errors, worksheet)

        else:
            registry.add_submittables(metadata_entities)

    def _register_module(self, metadata_entities, module_field_name, registry, workbook_errors, worksheet):
        removed_data = registry.add_modules(module_field_name, metadata_entities)
        workbook_errors.extend(self.list_data_removal_errors(worksheet.title, removed_data))

    def _import_modules(self, registry, workbook_errors):
        try:
            registry.import_modules()
        except LinkToConcreteEntityNotFound as error:
            location = error.module_entity.get_spreadsheet_location()
            workbook_errors.append({
                'location': f'sheet={location["worksheet_title"]} row_num={location["row_index"]}',
                'type': error.__class__.__name__, "detail": str(error)
            })

    def validate_worksheets(self, is_update, importable_worksheets):
        worksheets_with_uuid = []
        worksheets_without_uuid = []

        for worksheet in importable_worksheets:
            if worksheet.is_module_tab():
                continue
            concrete_type = self.template_mgr.get_concrete_type(worksheet.title)
            uuid_column = f'{concrete_type}.uuid'
            if worksheet.has_column(uuid_column):
                worksheets_with_uuid.append(worksheet.title)
            else:
                worksheets_without_uuid.append(worksheet.title)

        if is_update:
            errors = [MissingEntityUUIDFound(sheet_name) for sheet_name in worksheets_without_uuid]
        else:
            errors = [UnexpectedEntityUUIDFound(sheet_name) for sheet_name in worksheets_with_uuid]

        return [{"location": f'sheet={error.sheet_name}', "type": error.__class__.__name__, "detail": str(error)}
                for error in errors]

    def sheet_in_schemas(self, worksheet):
        schemas = self.template_mgr.template.json_schemas
        try:
            concrete_type = self.template_mgr.get_concrete_type(worksheet.title)
        except UnknownKeySchemaException as e:
            raise SheetNotFoundInSchemas(worksheet.title)
        module_field_name = worksheet.get_module_field_name()
        for schema in schemas:
            if 'name' in schema or 'title' in schema:
                schema_name = schema['name'] if 'name' in schema else schema['title']
                if schema_name == concrete_type:
                    if not worksheet.is_module_tab() or module_field_name in schema['properties']:
                        return True
                    raise SheetNotFoundInSchemas(worksheet.title)
        raise SheetNotFoundInSchemas(worksheet.title)

    @staticmethod
    def list_data_removal_errors(sheet, removed_data):
        errors = []
        for data in removed_data:
            e = DataRemoval(data['key'], data['value'])
            errors.append({"location": f'sheet={sheet}', "type": e.__class__.__name__, "detail": str(e)})
        return errors


class WorksheetImporter:
    KEY_HEADER_ROW_IDX = 4
    USER_FRIENDLY_HEADER_ROW_IDX = 2
    START_ROW_OFFSET = 5

    UNKNOWN_ID_PREFIX = '_unknown_'

    def __init__(self, template: TemplateManager):
        self.template = template
        self.unknown_id_ctr = 0
        self.logger = logging.getLogger(__name__)
        self.concrete_entity = None

    def do_import(self, ingest_worksheet: IngestWorksheet):
        records = []
        worksheet_errors = []
        try:
            row_template = self.template.create_row_template(ingest_worksheet)
            rows = ingest_worksheet.get_data_rows()
            for index, row in enumerate(rows):
                metadata, row_errors = row_template.do_import(row, ingest_worksheet.is_module_tab())
                for error in row_errors:
                    if 'location' in error:
                        error["location"] = f'sheet={ingest_worksheet.title} row={index}, {error["location"]}'
                    else:
                        error["location"] = f'sheet={ingest_worksheet.title} row={index}'
                    worksheet_errors.append(error)
                if not metadata.object_id:
                    metadata.object_id = self._generate_id()
                records.append(metadata)
        except Exception as e:
            worksheet_errors.append({
                "location": f'sheet={ingest_worksheet.title}',
                "type": e.__class__.__name__,
                "detail": str(e)
            })
        return records, worksheet_errors

    def _generate_id(self):
        self.unknown_id_ctr = self.unknown_id_ctr + 1
        return f'{self.UNKNOWN_ID_PREFIX}{self.unknown_id_ctr}'


class MultipleProjectsFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should only be associated to a single project.'
        super(MultipleProjectsFound, self).__init__(message)


class NoProjectFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should be associated to a project.'
        super(NoProjectFound, self).__init__(message)


class SheetNotFoundInSchemas(Exception):
    def __init__(self, sheet):
        message = f'The sheet named {sheet} was not found in the schema list.'
        super(SheetNotFoundInSchemas, self).__init__(message)
        self.sheet = sheet


class DataRemoval(Exception):
    def __init__(self, key, value):
        message = f'The column header [{key}] was not recognised, the following data has been removed: {value}.'
        super(DataRemoval, self).__init__(message)
        self.key = key
        self.value = value


class SchemaRetrievalError(Exception):
    pass


class UnexpectedEntityUUIDFound(Exception):
    def __init__(self, sheet_name):
        message = f'The {sheet_name} entities in the spreadsheet shouldn’t have UUIDs.'
        super(UnexpectedEntityUUIDFound, self).__init__(message)
        self.sheet_name = sheet_name


class MissingEntityUUIDFound(Exception):
    def __init__(self, sheet_name):
        message = f'The {sheet_name} entities in the spreadsheet should have UUIDs.'
        super(MissingEntityUUIDFound, self).__init__(message)
        self.sheet_name = sheet_name


class LinkToConcreteEntityNotFound(Exception):
    def __init__(self, module_entity: MetadataEntity):
        message = f'The module_entity is not linked to any concrete entity'
        super(LinkToConcreteEntityNotFound, self).__init__(message)
        self.module_entity = module_entity



import json
from typing import Dict

import jsonref
from jsonref import JsonLoader
from mergedeep import merge

from .descriptor import ComplexPropertyDescriptor


class SchemaParser:
    json_loader = JsonLoader()

    def __init__(self, json_schema,
                 ignored_properties=["required_properties", "describedBy", "schema_version", "schema_type",
                                     "provenance"]):
        self.ignored_properties = ignored_properties
        self.schema_descriptor = json_schema
        self.schema_dictionary = self.schema_descriptor

    @property
    def schema_descriptor(self):
        return self._schema_descriptor

    @schema_descriptor.setter
    def schema_descriptor(self, json_schema):
        """
        Given a json-formatted metadata schema, loads it into a Descriptor class which captures the structure as a
        dictionary and stores it as a private variable.

        :param json_schema: A raw metadata schema JSON object.
        """

        derefed_json_schema = self._deref_json_schema(json_schema)
        # merging the derefed json schema into the actual json schema. This preserves fields in the schema which
        # were specified alongside the $ref
        metadata_schema_data = merge(json_schema, derefed_json_schema)

        self._schema_descriptor = ComplexPropertyDescriptor(metadata_schema_data)

    @property
    def schema_dictionary(self):
        return self._schema_dictionary

    @schema_dictionary.setter
    def schema_dictionary(self, descriptor):
        """
        Given a Descriptor object, computes a dictionary representation describing the metadata schema with
        post-processing to removed
        ignored properties.

        :param descriptor: A Descriptor Object derived from a metadata schema JSON object.
        """
        self._schema_dictionary = self._get_schema_dictionary_with_ignored_fields_removed(
            descriptor.get_dictionary_representation_of_descriptor())

    def _deref_json_schema(self, json_schema: Dict) -> Dict:
        return jsonref.loads(json.dumps(json_schema), loader=SchemaParser.json_loader)

    def _get_schema_dictionary_with_ignored_fields_removed(self, dictionary_descriptor):
        """ Recursively removes all ignored properties in the given dictionary representation of a Descriptor which
        describes a metadata schema or a field within the metadata schema.

        :param dictionary_descriptor: A Descriptor in dictionary format from which to remove ignored properties.
        :return: The same dictionary representation of a Descriptor with all the properties listed in
                 self.ignored_properties removed.
        """

        in_post_processing_dictionary_descriptor = dictionary_descriptor

        for ignored_property in self.ignored_properties:
            if ignored_property in in_post_processing_dictionary_descriptor.keys():
                del in_post_processing_dictionary_descriptor[ignored_property]

        for key, value in in_post_processing_dictionary_descriptor.items():
            if isinstance(value, dict):
                post_processed_sub_dictionary_descriptor = \
                    self._get_schema_dictionary_with_ignored_fields_removed(
                        value)
                in_post_processing_dictionary_descriptor[key] = post_processed_sub_dictionary_descriptor

        return in_post_processing_dictionary_descriptor

    def get_map_of_paths_by_property_label(self, dictionary_descriptor):
        """
        Given a dictionary of Descriptor dictionaries by the module name for which they represent, returns a
        dictionary that maps the path via schema modules to each property that exists in the schemas. Each property
        is represented up to two times: once where the key is the user friendly name of the property and once as the
        path itself.

        :param dictionary_descriptor: A dictionary where each key is a module name and the value is a dictionary
        representation of its respective Descriptor object.
        """

        label_map = {}

        for metadata_schema, metadata_schema_properties in dictionary_descriptor.items():
            self._add_paths_to_map(metadata_schema_properties, label_map, metadata_schema)
        return label_map

    def _add_paths_to_map(self, metadata_property_dictionary, current_label_map, path_so_far):
        for property_key, property_value in metadata_property_dictionary.items():
            # Only put values into the map that are not metadata about the schema itself and not about the uuid.
            if isinstance(property_value, dict) and property_key != "schema" and property_key != "uuid":
                fully_qualified_property_label = path_so_far + "." + property_key
                current_label_map = self._put_into_map(fully_qualified_property_label, fully_qualified_property_label,
                                                       current_label_map)
                if "user_friendly" in property_value.keys():
                    user_friendly_property_label = property_value["user_friendly"]
                    current_label_map = self._put_into_map(user_friendly_property_label, fully_qualified_property_label,
                                                           current_label_map)

                self._add_paths_to_map(property_value, current_label_map, fully_qualified_property_label)
        return current_label_map

    def get_tab_representation_of_schema(self):
        """
        Returns a dictionary representing the way the schema would look as part of a tab in a spreadsheet where each
        of its properties (including embedded properties) are all flattened to be column names.
        """

        tab_key = self.schema_descriptor.get_schema_module_name()
        tab_display_name = tab_key[0].upper() + tab_key[1:].replace("_", " ")
        return {tab_key: {"display_name": tab_display_name,
                          "columns": self._get_columns_names_for_metadata_schema(tab_key, self.schema_dictionary)}}

    def _get_columns_names_for_metadata_schema(self, root_schema_name, root_schema_dictionary):
        list_of_column_names = []
        for key, value in root_schema_dictionary.items():
            if isinstance(value, dict) and key != "schema":
                next_root_schema_name = root_schema_name + "." + key
                children_column_names = self._get_columns_names_for_metadata_schema(next_root_schema_name, value)
                if children_column_names:
                    list_of_column_names += children_column_names
                else:
                    list_of_column_names.append(next_root_schema_name)

        return list_of_column_names

    @staticmethod
    def _put_into_map(key, value, current_label_map):
        key = key.lower()
        value = value.lower()

        if key in current_label_map.keys():
            current_values = current_label_map[key]
            current_values.append(value)
            current_label_map[key] = current_values
        else:
            current_label_map[key] = [value]

        return current_label_map



#!/usr/bin/env python3

from hca_ingest.api.ingestapi import IngestApi
from dataclasses import dataclass
import json
import datetime
import argparse
import uuid


class EntityMap:
    def __init__(self):
        self.uuid_map = dict()
        self.uri_map = dict()

    def add(self, entities: list, uuid_path: list = None):
        for e in entities:
            uri = e["_links"]["self"]["href"]
            self.uri_map[uri] = e
            if uuid_path:
                uuid = EntityMap.value_at_path(uuid_path, e)
            else:
                uuid = e["uuid"]["uuid"]
            self.uuid_map[uuid] = e

    def uuids(self):
        return list(self.uuid_map.keys())

    def uris(self):
        return list(self.uri_map.keys())

    def ids(self, with_uris=False):
        return self.uris() if with_uris else self.uuids()

    @staticmethod
    def from_list(entities: list, uuid_path: list = None):
        e = EntityMap()
        e.add(entities, uuid_path)
        return e

    @staticmethod
    def value_at_path(path, data):
        return EntityMap.value_at_path(path[1:], data[path[0]]) if len(path) > 0 else data

    def __len__(self):
        return len(self.uuid_map)

    def __iter__(self):
        return list(self.uuid_map.values()).__iter__()


@dataclass
class EntityRedaction:
    sequence_files: EntityMap = EntityMap()
    supplementary_files: EntityMap = EntityMap()
    biomaterials: EntityMap = EntityMap()
    processes: EntityMap = EntityMap()
    protocols: EntityMap = EntityMap()
    projects: EntityMap = EntityMap()
    primary_bundle_manifests: EntityMap = EntityMap()
    primary_submissions: EntityMap = EntityMap()

    analysis_processes: EntityMap = EntityMap()
    analysis_protocols: EntityMap = EntityMap()
    analysis_files: EntityMap = EntityMap()
    analysis_bundle_manifests: EntityMap = EntityMap()
    analysis_submissions: EntityMap = EntityMap()

    def to_redaction_manifest(self, with_uris=False) -> dict:
        return {
            "sequence_files": self.sequence_files.ids(with_uris),
            "supplementary_files": self.supplementary_files.ids(with_uris),
            "biomaterials": self.biomaterials.ids(with_uris),
            "processes": self.processes.ids(with_uris),
            "protocols": self.protocols.ids(with_uris),
            "projects": self.projects.ids(with_uris),
            "primary_bundle_manifests": self.primary_bundle_manifests.ids(with_uris),
            "primary_submissions": self.primary_submissions.ids(with_uris),

            "analysis_processes": self.analysis_processes.ids(with_uris),
            "analysis_protocols": self.analysis_protocols.ids(with_uris),
            "analysis_files": self.analysis_files.ids(with_uris),
            "analysis_bundle_manifests": self.analysis_bundle_manifests.ids(with_uris),
            "analysis_submissions": self.analysis_submissions.ids(with_uris)
        }


class RedactionManifestUtils:

    @staticmethod
    def create_redaction_manifest(project_uuid: str, ingest_url: str, with_uris=False):
        entity_redaction = RedactionManifestUtils.generate_entity_redaction(project_uuid, ingest_url)
        redaction_manifest = entity_redaction.to_redaction_manifest(with_uris)

        redaction_manifest_filename = f'redact_{project_uuid}_{datetime.datetime.utcnow().isoformat()}.json'
        with open(redaction_manifest_filename, "w") as f:
            print(f'Writing redaction manifest to {redaction_manifest_filename}\n')
            json.dump(redaction_manifest, f)

        print("Finished writing redaction manifest to {redaction_manifest_filename}\n")

    @staticmethod
    def generate_entity_redaction(project_uuid: str, ingest_url: str) -> 'EntityRedaction':
        ingest_client = IngestApi(ingest_url)

        project = ingest_client.get_project_by_uuid(project_uuid)
        submission_for_project = list(ingest_client.get_related_entities("submissionEnvelopes", project, "submissionEnvelopes"))[0]
        submissions_for_project = EntityMap.from_list([submission_for_project])

        files = list(ingest_client.get_related_entities("files", submission_for_project, "files"))
        supplementary_files = EntityMap.from_list([f for f in files if "supplementary_file" in f["content"]["describedBy"]])
        sequence_files = EntityMap.from_list([f for f in files if "sequence_file" in f["content"]["describedBy"]])

        assert len(supplementary_files) + len(sequence_files) == len(files)

        biomaterials = EntityMap.from_list(ingest_client.get_related_entities("biomaterials", submission_for_project, "biomaterials"))
        processes = EntityMap.from_list(ingest_client.get_related_entities("processes", submission_for_project, "processes"))
        protocols = EntityMap.from_list(ingest_client.get_related_entities("protocols", submission_for_project, "protocols"))
        projects = EntityMap.from_list(ingest_client.get_related_entities("projects", submission_for_project, "projects"))
        primary_bundle_manifests = EntityMap.from_list(ingest_client.get_related_entities("bundleManifests", submission_for_project, "bundleManifests"), ["bundleUuid"])

        analysis_processes = EntityMap()
        analysis_protocols = EntityMap()
        analysis_files = EntityMap()
        analysis_submissions = EntityMap()
        analysis_bundle_manifests = EntityMap()

        for file in sequence_files:
            input_to_processes = ingest_client.get_related_entities("inputToProcesses", file, "processes")
            for analysis_process in input_to_processes:
                analysis_processes.add([analysis_process])
                analysis_protocols.add(ingest_client.get_related_entities("protocols", analysis_process, "protocols"))
                analysis_files.add(ingest_client.get_related_entities("derivedFiles", analysis_process, "files"))
                analysis_submission = list(ingest_client.get_related_entities("submissionEnvelopes", analysis_process, "submissionEnvelopes"))[0]
                analysis_submissions.add([analysis_submission])
                analysis_bundle_manifests.add(ingest_client.get_related_entities("bundleManifests", analysis_submission, "bundleManifests"), ["bundleUuid"])

        assert len(projects.uuids()) == 1

        return EntityRedaction(sequence_files, supplementary_files, biomaterials, processes, protocols, projects,
                               primary_bundle_manifests, submissions_for_project, analysis_processes,
                               analysis_protocols, analysis_files, analysis_bundle_manifests, analysis_submissions)

    @staticmethod
    def envs():
        return {
            "dev": "https://api.ingest.dev.data.humancellatlas.org",
            "integration": "https://api.ingest.integration.data.humancellatlas.org",
            "staging": "https://api.ingest.staging.data.humancellatlas.org",
            "prod": "https://api.ingest.data.humancellatlas.org"
        }

    @staticmethod
    def ingest_api_url(env: str):
        return RedactionManifestUtils.envs()[env]


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate a redaction manifest for the HCA DCP ingestion service.\n "
                                                 "Given a project UUID, the redaction manifest will specify the "
                                                 "entities in the project to be redacted.")

    parser.add_argument('project-uuid', type=str, help='UUID of project to redact')
    parser.add_argument('--env', type=str, help='Environment to target')
    parser.add_argument('--ingest-url', type=int, help='URL of the ingest API')

    args = parser.parse_args()

    project_uuid = args.project_uuid
    env = args.env
    ingest_url = args.ingest_url

    try:
        uuid.UUID(project_uuid)
    except Exception:
        parser.error(f'{project_uuid} is not a valid UUID')

    if not env and not ingest_url:
        parser.error("Either --env or --ingest-url must be provided")

    if env not in RedactionManifestUtils.envs():
        parser.error(f'provided env "{env}" not in {list(RedactionManifestUtils.envs().keys())}')

    if not ingest_url:
        ingest_url = RedactionManifestUtils.ingest_api_url(env)

    RedactionManifestUtils.create_redaction_manifest(project_uuid, ingest_url, with_uris=True)
    print("Done\n")



import os

from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)
install_requires = [line.rstrip() for line in open(os.path.join(base_dir, 'requirements.in'))]

with open("README.md", "r", encoding="utf-8") as md:
    long_description = md.read()

setup(
    name='hca-ingest',
    version='2.5.0',
    description='A library to communicate with the Human Cell Atlas ingest API hosted by the EBI for creation and management of HCA project submissions.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ebi-ait/ingest-client",
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    include_package_data=True
)



class ValidationReport:
    def __init__(self, validation_state="", error_reports=list()):
        self.validation_state = validation_state
        self.error_reports = error_reports  # list of ErrorReport

    def errors_to_dict(self):
        return [error.to_dict() for error in self.error_reports]

    @staticmethod
    def validation_report_ok():
        report = ValidationReport()
        report.validation_state = "VALID"
        return report



from abc import abstractmethod
from enum import Enum

from hca_ingest.importer.conversion.exceptions import InvalidBooleanValue


class DataType(Enum):
    STRING = 'string'
    INTEGER = 'integer'
    BOOLEAN = 'boolean'
    NUMBER = 'number'
    UNDEFINED = 'undefined'

    @staticmethod
    def find(value: str):
        data_type = DataType.UNDEFINED
        if value is not None:
            try:
                data_type = DataType(value.lower())
            except ValueError:
                pass
        return data_type


class Converter:

    @abstractmethod
    def convert(self, data):
        raise NotImplementedError()

    @abstractmethod
    def type(self):
        raise NotImplementedError()


class DefaultConverter(Converter):

    def convert(self, data):
        return data

    def type(self):
        return DataType.UNDEFINED


class StringConverter(Converter):

    def convert(self, data):
        return str(data).strip()

    def type(self):
        return DataType.STRING


class IntegerConverter(Converter):

    def convert(self, data):
        try:
            return int(data)
        except ValueError:
            return int(float(data))

    def type(self):
        return DataType.INTEGER

class NumberConverter(Converter):

    def convert(self, data):
        return float(data)

    def type(self):
        return DataType.NUMBER


BOOLEAN_TABLE = {
    'true': True,
    'yes': True,
    'false': False,
    'no': False
}


class BooleanConverter(Converter):

    def convert(self, data):
        value = BOOLEAN_TABLE.get(data.lower())
        if value is None:
            raise InvalidBooleanValue(data)
        return value

    def type(self):
        return DataType.BOOLEAN


CONVERTER_MAP = {
    DataType.STRING: StringConverter(),
    DataType.INTEGER: IntegerConverter(),
    DataType.BOOLEAN: BooleanConverter(),
    DataType.NUMBER: NumberConverter()

}


class ListConverter(Converter):

    def __init__(self, data_type: DataType = DataType.STRING, base_converter: Converter = None):
        self.base_type = data_type
        if base_converter is not None:
            self.base_type = base_converter.type()
            self.base_converter = base_converter
        else:
            self.base_converter = CONVERTER_MAP.get(data_type, CONVERTER_MAP[DataType.STRING])

    def convert(self, data):
        data = str(data)
        value = data.split('||')
        value = [self.base_converter.convert(elem) for elem in value]
        return value

    def type(self):
        return self.base_type


DEFAULT = DefaultConverter()



from typing import Dict

from hca_ingest.api.ingestapi import IngestApi
from .entity import Entity


class DataCollector:
    def __init__(self, ingest_api: IngestApi):
        self.api = ingest_api

    def collect_data_by_submission_uuid(self, submission_uuid) -> Dict[str, Entity]:
        submission = self.api.get_submission_by_uuid(submission_uuid)
        entity_dict = self.__build_entity_dict(submission)
        return entity_dict

    def __build_entity_dict(self, submission):
        data_by_submission = self.__get_submission_data(submission)
        entity_dict = {}
        for entity_json in data_by_submission:
            entity = Entity.from_json(entity_json)
            entity_dict[entity.id] = entity
        linking_map = self.__get_linking_map(submission)
        self.__set_inputs(entity_dict, linking_map)
        return entity_dict

    def __get_submission_data(self, submission):
        submission_id = submission['_links']['self']['href'].split('/')[-1]
        project_json = self.api.get_related_project(submission_id)
        if project_json:
            submission_data = [
                project_json
            ]
        else:
            raise Exception('There should be a project')

        self.__get_entities_by_submission_and_type(submission_data, submission, 'biomaterials')
        self.__get_entities_by_submission_and_type(submission_data, submission, 'processes')
        self.__get_entities_by_submission_and_type(submission_data, submission, 'protocols')
        self.__get_entities_by_submission_and_type(submission_data, submission, 'files')

        return submission_data

    def __get_linking_map(self, submission):
        linking_map_url = submission['_links']['linkingMap']['href']
        headers = {'Content-type': 'application/json', 'Accept': 'application/hal+json'}
        r = self.api.get(linking_map_url, headers=headers)
        r.raise_for_status()
        linking_map = r.json()
        return linking_map

    @staticmethod
    def __set_inputs(entity_dict, linking_map):
        entities_with_inputs = list(linking_map['biomaterials'].keys()) + list(
            linking_map['files'].keys())

        for entity_id in entities_with_inputs:
            entity = entity_dict[entity_id]
            entity_link = linking_map[entity.domain_type + 's'][entity.id]
            derived_by_processes = entity_link.get('derivedByProcesses')

            if derived_by_processes and len(derived_by_processes) > 0:
                # Check if derivedByProcesses returns more than 1
                # It shouldn't happen because it's not possible to do it via spreadsheet
                if len(derived_by_processes) > 1:
                    raise ValueError(
                        f'The {entity.concrete_type} with {entity.uuid} has more than one processes which derived it')

                process_id = entity_link['derivedByProcesses'][0]
                protocol_ids = linking_map['processes'][process_id]['protocols']
                input_biomaterial_ids = linking_map['processes'][process_id]['inputBiomaterials']
                input_files_ids = linking_map['processes'][process_id]['inputFiles']

                process = entity_dict[process_id]
                protocols = [entity_dict[protocol_id] for protocol_id in protocol_ids]
                input_biomaterials = [entity_dict[id] for id in input_biomaterial_ids]
                input_files = [entity_dict[id] for id in input_files_ids]

                entity.set_input(input_biomaterials, input_files, process, protocols)

    def __get_entities_by_submission_and_type(self, data_by_submission, submission, entity_type):
        entity_json = \
            self.api.get_related_entities(entity_type, submission, entity_type)
        if entity_json:
            data_by_submission.extend(list(entity_json))
